# Copyright 2021 The KCL Authors. All rights reserved.

from typing import Optional, Any, List

import kclvm.kcl.error as kcl_error
import kclvm.kcl.ast as ast
import kclvm.compiler.astutil as astutil
import kclvm.compiler.parser as parser
import kclvm.compiler.extension.builtin.system_module.json as json
import kclvm.compiler.extension.builtin.system_module.yaml as yaml
import kclvm.program.eval as eval


class ValidationMeta:
    TEMP_FILE = "validationTempKCLCode.k"


class ValidationDataFormat:
    """
    KCL validation data formats including yaml and json.
    """

    YAML: str = "yaml"
    JSON: str = "json"
    MAPPING = {
        "YAML": yaml.KMANGLED_decode,
        "JSON": json.KMANGLED_decode,
    }


def validate_code(
    data: str,
    code: str,
    schema: Optional[str] = None,
    attribute_name: str = "value",
    format: str = "json",
    filename: str = None,
) -> bool:
    """Validate the data string using the schema code string, when the parameter
    `schema` is omitted, use the first scheam appeared in the code, when the schema
    if not found, raise an schema not found error.

    Returns a bool result denoting whether validating success, raise an error
    when validating failed because of the file not found error, schema not found
    error, syntax error, check error, etc.

    Parameters
    ----------
    data : str
        A JSON or YAML data string.
    code : str
        A KCL code string.
    schema : str
        The schema name required for verification.
    attribute_name : str
        The validation attribute name, default is `value`.
    format: str, default is "json"
        The data format, suppored json, JSON, yaml and YAML.
    filename: str, default is None
        The filename of the KCL code.

    Examples
    --------
    >>> data = '{"key": "value"}'  # A JSON data string
    >>> code = '''
        schema Person:
            key: str

            check:
                "value" in key  # `key` is required and `key` must contain "value"
        '''
    >>> validate_code(data, code)
    True

    """
    check_validation_para(data, code, format)
    # 1. Parse kcl code string to the AST module.
    module = parser.ParseFile(filename=filename or ValidationMeta.TEMP_FILE, code=code)
    schema_list = astutil.filter_stmt(module, ast.SchemaStmt)
    # 2. Deserialize data str and covert it to a KCL AST node.
    decoder = ValidationDataFormat.MAPPING.get(format.upper())
    value = decoder(data)
    schema_name = schema or (schema_list[0].name if schema_list else None)
    node_list = validate_value_to_ast_node_list(value, schema_name, attribute_name)
    # 3. Insert the value AST node into the module and eval
    module.body = node_list + module.body
    eval.EvalAST(module)
    return True


def validate_code_with_attr_data(
    data: str,
    code: str,
    schema: Optional[str] = None,
    format: str = "json",
) -> bool:
    """Validate the data string using the schema code string, when the parameter
    `schema` is omitted, use the first scheam appeared in the code, when the schema
    if not found, raise an schema not found error.

    Returns a bool result denoting whether validating success, raise an error
    when validating failed because of the file not found error, schema not found
    error, syntax error, check error, etc.

    Parameters
    ----------
    data : str
        A JSON or YAML data string including the attribute key
    code : str
        A KCL code string.
    schema : str
        The schema name required for verification.
    format: str, default is "json"
        The data format, suppored json, JSON, yaml and YAML.

    Examples
    --------
    >>> data = '{"attr": {"key": "value"}}'  # A JSON data string including the attribute name
    >>> code = '''
        schema Person:
            key: str

            check:
                "value" in key  # `key` is required and `key` must contain "value"
        '''
    >>> validate_code_with_attr_data(data, code)
    True
    """
    check_validation_para(data, code, format)
    decoder = ValidationDataFormat.MAPPING.get(format.upper())
    value = decoder(data)
    if not value or not isinstance(value, dict) or len(value) != 1:
        raise ValueError(
            f"Invalid parameter data: {data}, expected a dict with only one attribute"
        )
    attribute_name = list(value.keys())[0]
    data = json.KMANGLED_encode(value[attribute_name])
    return validate_code(
        data=data,
        code=code,
        schema=schema,
        attribute_name=attribute_name,
        format=format,
    )


def check_validation_para(data: str, code: str, format: str):
    if data is None or not isinstance(data, str):
        raise ValueError(f"Invalid parameter data: {data}")
    if code is None or not isinstance(code, str):
        raise ValueError(f"Invalid parameter code: {code}")
    if (
        format is None
        or not isinstance(format, str)
        or format.upper() not in ValidationDataFormat.MAPPING
    ):
        raise ValueError(
            f"Invalid parameter format: {format}, expected one of {ValidationDataFormat.MAPPING.keys()}"
        )


def validate_value_to_ast_node_list(
    value: Any, schema_name: str = None, attribute_name: str = "value"
) -> List[ast.AST]:
    """Covert a validation value to a KCL AST node"""

    def build_assign_node(attribute_name: str, node: ast.AST) -> List[ast.AssignStmt]:
        if not attribute_name:
            raise ValueError(f"Invalid parameter attribute_name: {attribute_name}")
        assign_stmt = ast.AssignStmt()
        assign_stmt.value = node
        assign_stmt.targets = [
            ast.Identifier(
                names=[attribute_name],
                ctx=ast.ExprContext.STORE,
            )
        ]
        return [assign_stmt]

    if isinstance(value, (int, float, bool, str, list, tuple, set, dict)):
        node = value_to_ast(value, schema_name)
        return build_assign_node(attribute_name, node)
    else:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CompileError_TYPE,
            arg_msg=f"invalid validation data value {value}",
        )


def value_to_ast(value: Any, schema_name: Optional[str] = None) -> ast.AST:
    node = None
    if value is None:
        node = ast.NameConstantLit(value=None)
    elif isinstance(value, (list, tuple, set)):
        node = ast.ListExpr()
        node.elts = [value_to_ast(v, schema_name) for v in value]
    elif isinstance(value, dict):
        config = ast.ConfigExpr()
        if schema_name:
            node = ast.SchemaExpr()
            node.name = ast.Identifier(names=[schema_name])
            node.config = config
        else:
            node = config
        for k, v in value.items():
            config.items.append(
                ast.ConfigEntry(
                    key=value_to_ast(k, schema_name),
                    value=value_to_ast(v),
                )
            )
    elif isinstance(value, (bool, int, float, str)):
        node = astutil.BuildLitNodeFromValue(value)
    return node



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gpyrpc/protorpc_wire.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1agpyrpc/protorpc_wire.proto\x12\rprotorpc_wire"}\n\rRequestHeader\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x17\n\x0fraw_request_len\x18\x03 \x01(\r\x12%\n\x1dsnappy_compressed_request_len\x18\x04 \x01(\r\x12\x10\n\x08\x63hecksum\x18\x05 \x01(\r"\x7f\n\x0eResponseHeader\x12\n\n\x02id\x18\x01 \x01(\x04\x12\r\n\x05\x65rror\x18\x02 \x01(\t\x12\x18\n\x10raw_response_len\x18\x03 \x01(\r\x12&\n\x1esnappy_compressed_response_len\x18\x04 \x01(\r\x12\x10\n\x08\x63hecksum\x18\x05 \x01(\r*.\n\x05\x43onst\x12\x08\n\x04ZERO\x10\x00\x12\x1b\n\x16MAX_REQUEST_HEADER_LEN\x10\x80\x08\x62\x06proto3'
)

_CONST = DESCRIPTOR.enum_types_by_name["Const"]
Const = enum_type_wrapper.EnumTypeWrapper(_CONST)
ZERO = 0
MAX_REQUEST_HEADER_LEN = 1024


_REQUESTHEADER = DESCRIPTOR.message_types_by_name["RequestHeader"]
_RESPONSEHEADER = DESCRIPTOR.message_types_by_name["ResponseHeader"]
RequestHeader = _reflection.GeneratedProtocolMessageType(
    "RequestHeader",
    (_message.Message,),
    {
        "DESCRIPTOR": _REQUESTHEADER,
        "__module__": "gpyrpc.protorpc_wire_pb2"
        # @@protoc_insertion_point(class_scope:protorpc_wire.RequestHeader)
    },
)
_sym_db.RegisterMessage(RequestHeader)

ResponseHeader = _reflection.GeneratedProtocolMessageType(
    "ResponseHeader",
    (_message.Message,),
    {
        "DESCRIPTOR": _RESPONSEHEADER,
        "__module__": "gpyrpc.protorpc_wire_pb2"
        # @@protoc_insertion_point(class_scope:protorpc_wire.ResponseHeader)
    },
)
_sym_db.RegisterMessage(ResponseHeader)

if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _CONST._serialized_start = 301
    _CONST._serialized_end = 347
    _REQUESTHEADER._serialized_start = 45
    _REQUESTHEADER._serialized_end = 170
    _RESPONSEHEADER._serialized_start = 172
    _RESPONSEHEADER._serialized_end = 299
# @@protoc_insertion_point(module_scope)



# Copyright 2020 The KCL Authors. All rights reserved.

import kclvm.kcl.ast as ast


class TreeTransformer(ast.TreeWalker):
    """The TreeTransformer subclass that walks the abstract syntax tree and
    allows modification of nodes.

    The `TreeTransformer` will walk the AST and use the return value of the
    walker methods to replace or remove the old node.  If the return value of
    the walker method is ``None``, the node will be removed from its location,
    otherwise it is replaced with the return value.  The return value may be the
    original node in which case no replacement takes place.

    Keep in mind that if the node you're operating on has child nodes you must
    either transform the child nodes yourself or call the :meth:`generic_walk`
    method for the node first.

    For nodes that were part of a collection of statements (that applies to all
    statement nodes), the walker may also return a list of nodes rather than
    just a single node, e.g., `module` AST contains the `body` attribute, which
    is a list of statement

    Example:

        class ChangeSchemaExprNameTransformer(TreeTransformer):
            def walk_SchemaExpr(self, t: ast.SchemaExpr):
                if t.name.get_name() == 'Person':
                    t.name.set_name('PersonNew')
                return t

    Usually we can use the transformer like this::

       node = YourTransformer().walk(node)

    """

    # Base transformer functions

    def get_node_name(self, t: ast.AST):
        """Get the ast.AST node name"""
        assert isinstance(t, ast.AST)
        return t.type

    def generic_walk(self, t: ast.AST):
        """Called if no explicit walker function exists for a node."""
        for field, old_value in ast.iter_fields(t):
            if isinstance(old_value, list):
                new_values = []
                for value in old_value:
                    if isinstance(value, ast.AST):
                        value = self.walk(value)
                        if value is None:
                            continue
                        elif not isinstance(value, ast.AST):
                            new_values.extend(value)
                            continue
                    elif isinstance(value, list):
                        value = [self.walk(v) for v in value]
                    new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, ast.AST):
                new_node = self.walk(old_value)
                if new_node is None:
                    delattr(t, field)
                else:
                    setattr(t, field, new_node)
        return t



"""
The `kcl_err_msg` file mainly contains constants used in KCL error messages.

INT_OVER_FLOW_MSG: String constants ending with "_MSG" are the recommended string template when raising KCL exceptions.

ErrEwcode: All ewcodes for KCL exception.

ErrMsgContent_EN: A dict for kcl exception message,
                    <key: value> = <ewcode: message>

ErrArgMsgDefault_EN: A dict for default arguments provided when a kcl exception is raised,
                        <key: value> = <ewcode: arg>

ErrName_EN: A dict for kcl exception names,
                <key: value> = <ewcode: exception_name>

KCLErrMsgManager: A singleton class that encapsulates some methods of obtaining KCL exception information

:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""
import sys
from abc import ABCMeta, abstractmethod
from kclvm.internal.util import PreCheck, CheckRules, PostCheck
from kclvm.internal.util.check_utils import PostSimpleExprCheck

ERROR = "E"
WARNING = "W"

SYNTAX = "1"
COMPLIER = "2"
RUNTIME = "3"

ATTRIBUTE = "A"
SCHEMA = "B"
MIXIN = "C"
INHERIT = "D"
IMPORT = "F"
TYPE = "G"
DECORATOR = "H"
ARGUMENT = "I"
OVERFLOW = "K"
COMPLING = "L"
RUNNING = "M"
DEPRECATED = "N"
DOC = "P"
IMMUTABLE = "Q"

INT_OVER_FLOW_MSG = "{}: A {} bit integer overflow"
FLOAT_OVER_FLOW_MSG = "{}: A {}-bit floating point number overflow"
FLOAT_UNDER_FLOW_MSG = "{}: A {}-bit floating point number underflow"
DEPRECATED_WARNING_MSG = "{} was deprecated {}"

SCHEMA_CHECK_FILE_MSG_ERR = "Instance check failed"
SCHEMA_CHECK_FILE_MSG_COND = "Check failed on the condition"

UNIQUE_KEY_MSG = "Variable name '{}' must be unique in package context"
CANNOT_ADD_MEMBERS_MSG = "{}: No such member in the schema '{}'"
RECURSIVE_LOADING_MODULE_MSG = "In module {}, recursively loading modules: {}"
CANNOT_FIND_MODULE_MSG = "Cannot find the module {} from {}"
MULTI_INHERIT_MSG = "Multiple inheritance of {} is prohibited"
INDENTATION_ERROR_MSG = "Unindent {} does not match any outer indentation level"
UNKNOWN_DECORATOR_MSG = "UnKnown decorator {}"
INVALID_DECORATOR_TARGET_MSG = "Invalid decorator target {}"
NAME_ERROR_MSG = "Name error {}"
INVALID_FORMAT_SPEC_MSG = "{} is invalid format spec"


class ErrEwcode:
    KCLException_Ew = "00000"  # 00000
    KCLError_Ew = ERROR + "0000"  # E0000
    KCLWarning_Ew = WARNING + "0000"  # W0000

    KCLSyntaxException_Ew = "0" + SYNTAX + "000"  # 01000
    KCLCompileException_Ew = "0" + COMPLIER + "000"  # 02000
    KCLRuntimeException_Ew = "0" + RUNTIME + "000"  # 03000

    KCLAttributeException_Ew = "00" + ATTRIBUTE + "00"  # 00A00
    KCLSchemaException_Ew = "00" + SCHEMA + "00"  # 00B00
    KCLMixinException_Ew = "00" + MIXIN + "00"  # 00C00
    KCLInheritException_Ew = "00" + INHERIT + "00"  # 00D00
    KCLImportException_Ew = "00" + IMPORT + "00"  # 00F00
    KCLTypeException_Ew = "00" + TYPE + "00"  # 00G00
    KCLDecoratorException_Ew = "00" + DECORATOR + "00"  # 00H00
    KCLArgumentException_Ew = "00" + ARGUMENT + "00"  # 00I00
    KCLOverflowException_Ew = "00" + OVERFLOW + "00"  # 00K00
    KCLComplingException_Ew = "00" + COMPLING + "00"  # 00L00
    KCLRunningException_Ew = "00" + RUNNING + "00"  # 00M00
    KCLDeprecatedException_Ew = "00" + DEPRECATED + "00"  # 00N00
    KCLDocException_Ew = "00" + DOC + "00"  # 00P00
    KCLImmutableException_Ew = "00" + IMMUTABLE + "00"  # 00Q00

    InvalidSyntax_Ew = ERROR + SYNTAX + "001"  # E1001
    KCLTabError_Ew = ERROR + SYNTAX + "002"  # E1002
    KCLIndentationError_Ew = ERROR + SYNTAX + "003"  # E1003

    CannotFindModule_Ew = ERROR + COMPLIER + IMPORT + "04"  # E2F04
    FailedLoadModule_Ew = ERROR + COMPLIER + IMPORT + "05"  # E2F05
    RecursiveLoad_Ew = ERROR + COMPLIER + IMPORT + "06"  # E3F06

    FloatOverflow_Ew = ERROR + RUNTIME + OVERFLOW + "07"  # E3K04
    FloatUnderflow_Ew = WARNING + COMPLIER + OVERFLOW + "08"  # W2K08
    IntOverflow_Ew = ERROR + RUNTIME + OVERFLOW + "09"  # E3K09

    InvalidDocstring_Ew = WARNING + COMPLIER + DOC + "10"  # W2P10

    Deprecated_Ew = ERROR + RUNTIME + DEPRECATED + "11"  # E3N11
    DeprecatedWarning_Ew = WARNING + COMPLIER + DEPRECATED + "12"  # W2N12

    UnKnownDecorator_Ew = ERROR + COMPLIER + DECORATOR + "13"  # E2H13
    InvalidDecoratorTarget_Ew = ERROR + COMPLIER + DECORATOR + "14"  # E2H14

    MixinNamingError_Ew = ERROR + COMPLIER + MIXIN + "15"  # E2C15
    MixinStructureIllegal_Ew = ERROR + COMPLIER + MIXIN + "16"  # E2C16

    SchemaCheckFailure_Ew = ERROR + RUNTIME + SCHEMA + "17"  # E3B17
    CannotAddMembersComplieError_Ew = ERROR + COMPLIER + SCHEMA + "18"  # E2B17
    CannotAddMembersRuntimeError_Ew = ERROR + RUNTIME + SCHEMA + "19"  # E3B19
    IndexSignatureError_Ew = ERROR + COMPLIER + SCHEMA + "20"  # E2B20

    TypeRuntimeError_Ew = ERROR + RUNTIME + TYPE + "21"  # E3G21
    TypeComplieError_Ew = ERROR + COMPLIER + TYPE + "22"  # E2G22

    CompileError_Ew = ERROR + COMPLIER + COMPLING + "23"  # E2L23
    SelectorError_Ew = ERROR + COMPLIER + COMPLING + "24"  # E2L24
    KCLNameError_Ew = ERROR + COMPLIER + COMPLING + "25"  # E2L25
    KCLValueError_Ew = ERROR + COMPLIER + COMPLING + "26"  # E2L26
    KCLKeyError_Ew = ERROR + COMPLIER + COMPLING + "27"  # E2L27
    UniqueKeyError_Ew = ERROR + COMPLIER + COMPLING + "28"  # E2L28

    KCLAttributeComplieError_Ew = ERROR + COMPLIER + ATTRIBUTE + "29"  # E2A29
    KCLAttributeRuntimeError_Ew = ERROR + RUNTIME + ATTRIBUTE + "30"  # E3A30
    IllegalAttributeError_Ew = ERROR + COMPLIER + ATTRIBUTE + "31"  # E2A31

    MultiInheritError_Ew = ERROR + COMPLIER + INHERIT + "32"  # E2D32
    CycleInheritError_Ew = ERROR + COMPLIER + INHERIT + "33"  # E2D33
    IllegalInheritError_Ew = ERROR + COMPLIER + INHERIT + "34"  # E2D34

    IllegalArgumentRuntimeError_Ew = ERROR + RUNTIME + ARGUMENT + "35"  # E3I35
    IllegalArgumentComplieError_Ew = ERROR + COMPLIER + ARGUMENT + "36"  # E2I36
    IllegalArgumentSyntaxError_Ew = ERROR + SYNTAX + ARGUMENT + "37"  # E1I37

    EvaluationError_Ew = ERROR + RUNTIME + RUNNING + "38"  # E3M38
    InvalidFormatSpec_Ew = ERROR + RUNTIME + RUNNING + "39"  # E3M39
    KCLAssertionError_Ew = ERROR + RUNTIME + RUNNING + "40"  # E3M40

    ImmutableCompileError_Ew = ERROR + COMPLIER + COMPLING + "41"  # E3L41
    KCLRecursionError_Ew = ERROR + RUNTIME + RUNNING + "42"  # E3M42
    PlanError_Ew = ERROR + RUNTIME + RUNNING + "43"  # E3M43
    ImmutableRuntimeError_Ew = ERROR + RUNTIME + RUNNING + "44"  # E3M44

    @staticmethod
    def contains(ewcode: str):
        return ewcode in ErrEwcode.__dict__.values()


ErrMsgContent_EN: dict = {
    ErrEwcode.KCLException_Ew: "Exception",
    ErrEwcode.KCLError_Ew: "Error",
    ErrEwcode.KCLWarning_Ew: "Warning",
    ErrEwcode.KCLSyntaxException_Ew: "Syntax",
    ErrEwcode.KCLCompileException_Ew: "Compile",
    ErrEwcode.KCLRuntimeException_Ew: "Runtime",
    ErrEwcode.KCLAttributeException_Ew: "An attribute exception occurs",
    ErrEwcode.KCLSchemaException_Ew: "A schema exception occurs",
    ErrEwcode.KCLMixinException_Ew: "A mixin exception occurs",
    ErrEwcode.KCLInheritException_Ew: "An inherit exception occurs",
    ErrEwcode.KCLImportException_Ew: "An import exception occurs",
    ErrEwcode.KCLTypeException_Ew: "A type exception occurs",
    ErrEwcode.KCLDecoratorException_Ew: "A decorator exception occurs",
    ErrEwcode.KCLArgumentException_Ew: "An argument exception occurs",
    ErrEwcode.KCLOverflowException_Ew: "An overflow exception occurs",
    ErrEwcode.KCLComplingException_Ew: "An compling exception occurs",
    ErrEwcode.KCLRunningException_Ew: "An running exception occurs",
    ErrEwcode.KCLDeprecatedException_Ew: "A deprecated exception occurs",
    ErrEwcode.KCLDocException_Ew: "A doc exception occurs",
    ErrEwcode.KCLImmutableException_Ew: "A Immutable exception occurs",
    ErrEwcode.InvalidSyntax_Ew: "Invalid syntax",
    ErrEwcode.KCLTabError_Ew: "Tab Error",
    ErrEwcode.KCLIndentationError_Ew: "Indentation Error",
    ErrEwcode.CannotFindModule_Ew: "Cannot find the module",
    ErrEwcode.FailedLoadModule_Ew: "Failed to load module",
    ErrEwcode.RecursiveLoad_Ew: "Recursively loading module",
    ErrEwcode.FloatOverflow_Ew: "Float overflow",
    ErrEwcode.FloatUnderflow_Ew: "Float underflow",
    ErrEwcode.IntOverflow_Ew: "Integer overflow",
    ErrEwcode.InvalidDocstring_Ew: "Invalid docstring",
    ErrEwcode.Deprecated_Ew: "Deprecated error",
    ErrEwcode.DeprecatedWarning_Ew: "Deprecated warning",
    ErrEwcode.UnKnownDecorator_Ew: "UnKnown decorator",
    ErrEwcode.InvalidDecoratorTarget_Ew: "Invalid Decorator Target",
    ErrEwcode.MixinNamingError_Ew: "Illegal mixin naming",
    ErrEwcode.MixinStructureIllegal_Ew: "Illegal mixin structure",
    ErrEwcode.SchemaCheckFailure_Ew: "Schema check is failed to check condition",
    ErrEwcode.CannotAddMembersComplieError_Ew: "Cannot add members to a schema",
    ErrEwcode.CannotAddMembersRuntimeError_Ew: "Cannot add members to a schema",
    ErrEwcode.IndexSignatureError_Ew: "Invalid index signature",
    ErrEwcode.TypeRuntimeError_Ew: "The type got is inconsistent with the type expected",
    ErrEwcode.TypeComplieError_Ew: "The type got is inconsistent with the type expected",
    ErrEwcode.CompileError_Ew: "A complie error occurs during compiling",
    ErrEwcode.SelectorError_Ew: "Selector Error",
    ErrEwcode.KCLNameError_Ew: "Name Error",
    ErrEwcode.KCLValueError_Ew: "Value Error",
    ErrEwcode.KCLKeyError_Ew: "Key Error",
    ErrEwcode.UniqueKeyError_Ew: "Unique key error",
    ErrEwcode.KCLAttributeComplieError_Ew: "Attribute error occurs during compiling",
    ErrEwcode.KCLAttributeRuntimeError_Ew: "Attribute error occurs at runtime",
    ErrEwcode.IllegalAttributeError_Ew: "Illegal attribute",
    ErrEwcode.MultiInheritError_Ew: "Multiple inheritance is illegal",
    ErrEwcode.CycleInheritError_Ew: "Cycle Inheritance is illegal",
    ErrEwcode.IllegalInheritError_Ew: "Illegal inheritance",
    ErrEwcode.IllegalArgumentRuntimeError_Ew: "Illegal command line argument at runtime",
    ErrEwcode.IllegalArgumentComplieError_Ew: "Illegal argument during compiling",
    ErrEwcode.IllegalArgumentSyntaxError_Ew: "Illegal argument syntax",
    ErrEwcode.EvaluationError_Ew: "Evaluation failure",
    ErrEwcode.InvalidFormatSpec_Ew: "Invalid format specification",
    ErrEwcode.KCLAssertionError_Ew: "Assertion failure",
    ErrEwcode.ImmutableCompileError_Ew: "Immutable variable is modified during compiling",
    ErrEwcode.ImmutableRuntimeError_Ew: "Immutable variable is modified at runtime",
    ErrEwcode.KCLRecursionError_Ew: "Recursively reference",
    ErrEwcode.PlanError_Ew: "Plan Error",
}

ErrArgMsgDefault_EN: dict = {
    ErrEwcode.KCLException_Ew: "Exception",
    ErrEwcode.KCLError_Ew: "Error",
    ErrEwcode.KCLWarning_Ew: "Warning",
    ErrEwcode.KCLSyntaxException_Ew: "Syntax",
    ErrEwcode.KCLCompileException_Ew: "Complie",
    ErrEwcode.KCLRuntimeException_Ew: "Runtime",
    ErrEwcode.KCLAttributeException_Ew: "An attribute exception occurs",
    ErrEwcode.KCLSchemaException_Ew: "A schema exception occurs",
    ErrEwcode.KCLMixinException_Ew: "A mixin exception occurs",
    ErrEwcode.KCLInheritException_Ew: "An inherit exception occurs",
    ErrEwcode.KCLImportException_Ew: "An import exception occurs",
    ErrEwcode.KCLTypeException_Ew: "A type exception occurs",
    ErrEwcode.KCLDecoratorException_Ew: "A decorator exception occurs",
    ErrEwcode.KCLArgumentException_Ew: "An argument exception occurs",
    ErrEwcode.KCLOverflowException_Ew: "An overflow exception occurs",
    ErrEwcode.KCLComplingException_Ew: "An compling exception occurs",
    ErrEwcode.KCLRunningException_Ew: "An running exception occurs",
    ErrEwcode.KCLDeprecatedException_Ew: "A deprecated exception occurs",
    ErrEwcode.KCLDocException_Ew: "A doc exception occurs",
    ErrEwcode.KCLImmutableException_Ew: "A Immutable exception occurs",
    ErrEwcode.InvalidSyntax_Ew: "Invalid syntax",
    ErrEwcode.KCLTabError_Ew: "Inconsistent use of tabs and spaces in indentation",
    ErrEwcode.KCLIndentationError_Ew: "Indentation Error",
    ErrEwcode.CannotFindModule_Ew: "Cannot find the module",
    ErrEwcode.FailedLoadModule_Ew: "Failed to load module",
    ErrEwcode.RecursiveLoad_Ew: "Recursively loading module",
    ErrEwcode.FloatOverflow_Ew: "Float overflow",
    ErrEwcode.FloatUnderflow_Ew: "Float underflow",
    ErrEwcode.IntOverflow_Ew: "Integer overflow",
    ErrEwcode.InvalidDocstring_Ew: "Invalid docstring",
    ErrEwcode.Deprecated_Ew: "Deprecated error",
    ErrEwcode.DeprecatedWarning_Ew: "Deprecated warning",
    ErrEwcode.UnKnownDecorator_Ew: "UnKnown decorator",
    ErrEwcode.InvalidDecoratorTarget_Ew: "Invalid Decorator Target",
    ErrEwcode.MixinNamingError_Ew: "Illegal mixin naming",
    ErrEwcode.MixinStructureIllegal_Ew: "Illegal mixin structure",
    ErrEwcode.SchemaCheckFailure_Ew: "Check failed on check conditions",
    ErrEwcode.CannotAddMembersComplieError_Ew: "Cannot add members to a schema",
    ErrEwcode.CannotAddMembersRuntimeError_Ew: "Cannot add members to a schema",
    ErrEwcode.IndexSignatureError_Ew: "Invalid index signature",
    ErrEwcode.TypeRuntimeError_Ew: "The type got is inconsistent with the type expected",
    ErrEwcode.TypeComplieError_Ew: "The type got is inconsistent with the type expected",
    ErrEwcode.CompileError_Ew: "A complie error occurs during compiling",
    ErrEwcode.SelectorError_Ew: "Selector Error",
    ErrEwcode.KCLNameError_Ew: "Name Error",
    ErrEwcode.KCLValueError_Ew: "Value Error",
    ErrEwcode.KCLKeyError_Ew: "Key Error",
    ErrEwcode.UniqueKeyError_Ew: "Unique key error",
    ErrEwcode.KCLAttributeComplieError_Ew: "Attribute error occurs during compiling",
    ErrEwcode.KCLAttributeRuntimeError_Ew: "Attribute error occurs at runtime",
    ErrEwcode.IllegalAttributeError_Ew: "Illegal attribute",
    ErrEwcode.MultiInheritError_Ew: "Multiple inheritance is illegal",
    ErrEwcode.CycleInheritError_Ew: "Cycle Inheritance is illegal",
    ErrEwcode.IllegalInheritError_Ew: "Illegal inheritance",
    ErrEwcode.IllegalArgumentRuntimeError_Ew: "Illegal command line argument at runtime",
    ErrEwcode.IllegalArgumentComplieError_Ew: "Illegal argument during compiling",
    ErrEwcode.IllegalArgumentSyntaxError_Ew: "Illegal argument syntax",
    ErrEwcode.EvaluationError_Ew: "Evaluation failure",
    ErrEwcode.InvalidFormatSpec_Ew: "Invalid format specification",
    ErrEwcode.KCLAssertionError_Ew: "Assertion failure",
    ErrEwcode.ImmutableCompileError_Ew: "Immutable variable is modified during compiling",
    ErrEwcode.ImmutableRuntimeError_Ew: "Immutable variable is modified at runtime",
    ErrEwcode.KCLRecursionError_Ew: "maximum recursion depth exceeded",
    ErrEwcode.PlanError_Ew: "Plan Error",
}

ErrName_EN: dict = {
    ErrEwcode.KCLException_Ew: "KCLException",
    ErrEwcode.KCLError_Ew: "KCLError",
    ErrEwcode.KCLWarning_Ew: "KCLWarning",
    ErrEwcode.KCLSyntaxException_Ew: "KCLSyntaxException",
    ErrEwcode.KCLCompileException_Ew: "KCLCompileException",
    ErrEwcode.KCLRuntimeException_Ew: "KCLRuntimeException",
    ErrEwcode.KCLAttributeException_Ew: "KCLAttributeException",
    ErrEwcode.KCLSchemaException_Ew: "KCLSchemaException",
    ErrEwcode.KCLMixinException_Ew: "KCLMixinException",
    ErrEwcode.KCLInheritException_Ew: "KCLInheritException",
    ErrEwcode.KCLImportException_Ew: "KCLImportException",
    ErrEwcode.KCLTypeException_Ew: "KCLTypeException",
    ErrEwcode.KCLDecoratorException_Ew: "KCLDecoratorException",
    ErrEwcode.KCLArgumentException_Ew: "KCLArgumentException",
    ErrEwcode.KCLOverflowException_Ew: "KCLOverflowException",
    ErrEwcode.KCLComplingException_Ew: "KCLComplingException",
    ErrEwcode.KCLRunningException_Ew: "KCLRunningException",
    ErrEwcode.KCLDeprecatedException_Ew: "KCLDeprecatedException",
    ErrEwcode.KCLDocException_Ew: "KCLDocException",
    ErrEwcode.KCLImmutableException_Ew: "KCLImmutableException",
    ErrEwcode.InvalidSyntax_Ew: "InvalidSyntax",
    ErrEwcode.KCLTabError_Ew: "KCLTabError",
    ErrEwcode.KCLIndentationError_Ew: "KCLIndentationError",
    ErrEwcode.CannotFindModule_Ew: "CannotFindModule",
    ErrEwcode.FailedLoadModule_Ew: "FailedLoadModule",
    ErrEwcode.RecursiveLoad_Ew: "RecursiveLoad",
    ErrEwcode.FloatOverflow_Ew: "FloatOverflow",
    ErrEwcode.FloatUnderflow_Ew: "FloatUnderflow",
    ErrEwcode.IntOverflow_Ew: "IntOverflow",
    ErrEwcode.InvalidDocstring_Ew: "InvalidDocstring",
    ErrEwcode.Deprecated_Ew: "Deprecated",
    ErrEwcode.DeprecatedWarning_Ew: "DeprecatedWarning",
    ErrEwcode.UnKnownDecorator_Ew: "UnKnownDecorator",
    ErrEwcode.InvalidDecoratorTarget_Ew: "InvalidDecoratorTarget",
    ErrEwcode.MixinNamingError_Ew: "MixinNamingError",
    ErrEwcode.MixinStructureIllegal_Ew: "MixinStructureIllegal",
    ErrEwcode.SchemaCheckFailure_Ew: "SchemaCheckFailure",
    ErrEwcode.CannotAddMembersComplieError_Ew: "CannotAddMembersComplieError",
    ErrEwcode.CannotAddMembersRuntimeError_Ew: "CannotAddMembersRuntimeError",
    ErrEwcode.IndexSignatureError_Ew: "IndexSignatureError",
    ErrEwcode.TypeRuntimeError_Ew: "TypeRuntimeError",
    ErrEwcode.TypeComplieError_Ew: "TypeComplieError",
    ErrEwcode.CompileError_Ew: "CompileError",
    ErrEwcode.SelectorError_Ew: "SelectorError",
    ErrEwcode.KCLNameError_Ew: "KCLNameError",
    ErrEwcode.KCLValueError_Ew: "KCLValueError",
    ErrEwcode.KCLKeyError_Ew: "KCLKeyError",
    ErrEwcode.UniqueKeyError_Ew: "UniqueKeyError",
    ErrEwcode.KCLAttributeComplieError_Ew: "KCLAttributeComplieError",
    ErrEwcode.KCLAttributeRuntimeError_Ew: "KCLAttributeRuntimeError",
    ErrEwcode.IllegalAttributeError_Ew: "IllegalAttributeError",
    ErrEwcode.MultiInheritError_Ew: "MultiInheritError",
    ErrEwcode.CycleInheritError_Ew: "CycleInheritError",
    ErrEwcode.IllegalInheritError_Ew: "IllegalInheritError",
    ErrEwcode.IllegalArgumentRuntimeError_Ew: "IllegalArgumentRuntimeError",
    ErrEwcode.IllegalArgumentComplieError_Ew: "IllegalArgumentComplieError",
    ErrEwcode.IllegalArgumentSyntaxError_Ew: "IllegalArgumentSyntaxError",
    ErrEwcode.EvaluationError_Ew: "EvaluationError",
    ErrEwcode.InvalidFormatSpec_Ew: "InvalidFormatSpec",
    ErrEwcode.KCLAssertionError_Ew: "KCLAssertionError",
    ErrEwcode.ImmutableCompileError_Ew: "ImmutableCompileError",
    ErrEwcode.ImmutableRuntimeError_Ew: "ImmutableRuntimeError",
    ErrEwcode.KCLRecursionError_Ew: "KCLRecursionError",
    ErrEwcode.PlanError_Ew: "PlanError",
}


class ErrMsgLang:
    EN_US_UTF_8 = "c.utf8"


class ErrMsg(metaclass=ABCMeta):
    @abstractmethod
    def get_defaule_ewcode(self) -> str:
        pass

    @abstractmethod
    def get_err_cate_ewcode(self, err_id: str) -> str:
        pass

    @abstractmethod
    def get_err_type_ewcode(self, err_id: str) -> str:
        pass

    @abstractmethod
    def get_err_cate_msg_by_errid(self, err_id: str) -> str:
        pass

    @abstractmethod
    def get_err_type_msg_by_errid(self, err_id: str) -> str:
        pass

    @abstractmethod
    def get_err_msg_by_errid(self, err_id: str) -> str:
        pass

    @abstractmethod
    def get_defaule_arg_msg_by_errid(self, err_id: str) -> str:
        pass

    @abstractmethod
    def get_err_code_by_errid(self, err_id: str) -> str:
        pass

    @abstractmethod
    def is_warning(self, err_id: str) -> bool:
        pass

    @abstractmethod
    def is_syntax_error(self, err_id: str) -> bool:
        pass

    @abstractmethod
    def is_compiler_error(self, err_id: str) -> bool:
        pass

    @abstractmethod
    def is_runtime_error(self, err_id: str) -> bool:
        pass

    @abstractmethod
    def get_err_name_by_ewcode(self, err_id: str):
        pass


class KCLErrMsg_EN(ErrMsg):
    @PostCheck(lambda result: ErrEwcode.contains(result))
    def get_defaule_ewcode(self) -> str:
        return ErrEwcode.KCLException_Ew

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: ErrEwcode.contains(result))
    @PostSimpleExprCheck(
        (lambda inputs, result: result == inputs["err_id"][:1] + "0000"), ["err_id"]
    )
    def get_err_cate_ewcode(self, err_id: str) -> str:
        return err_id[:1] + "0000"

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: ErrEwcode.contains(result))
    @PostSimpleExprCheck(
        (lambda inputs, result: result == "0" + inputs["err_id"][1:2] + "000"),
        ["err_id"],
    )
    def get_err_type_ewcode(self, err_id: str) -> str:
        return "0" + err_id[1:2] + "000"

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: result in ErrMsgContent_EN.values())
    def get_err_cate_msg_by_errid(self, err_id: str) -> str:
        return ErrMsgContent_EN[self.get_err_cate_ewcode(err_id)]

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: result in ErrMsgContent_EN.values())
    def get_err_type_msg_by_errid(self, err_id: str) -> str:
        return ErrMsgContent_EN[self.get_err_type_ewcode(err_id)]

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: result in ErrMsgContent_EN.values())
    def get_err_msg_by_errid(self, err_id: str) -> str:
        return ErrMsgContent_EN[err_id]

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: result in ErrArgMsgDefault_EN.values())
    def get_defaule_arg_msg_by_errid(self, err_id: str) -> str:
        return ErrArgMsgDefault_EN[err_id]

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    @PostSimpleExprCheck(
        (lambda inputs, result: result == "[" + inputs["err_id"] + "]"), ["err_id"]
    )
    def get_err_code_by_errid(self, err_id: str) -> str:
        return "[" + err_id + "]"

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, bool))
    def is_warning(self, err_id: str) -> bool:
        return err_id[:1] == WARNING

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, bool))
    def is_syntax_error(self, err_id: str) -> bool:
        return err_id[1:2] == SYNTAX

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, bool))
    def is_compiler_error(self, err_id: str) -> bool:
        return err_id[1:2] == COMPLIER

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, bool))
    def is_runtime_error(self, err_id: str) -> bool:
        return err_id[1:2] == RUNTIME

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: result in ErrName_EN.values())
    def get_err_name_by_ewcode(self, err_id: str) -> str:
        return ErrName_EN[err_id]


class KCLErrMsgManager(ErrMsg):
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self, lang: ErrMsgLang = ErrMsgLang.EN_US_UTF_8):
        self.ALL_KCL_ERROR_MSGS: dict = {ErrMsgLang.EN_US_UTF_8: KCLErrMsg_EN()}
        self.lang = lang
        self.KCL_ERROR_MSG = self.ALL_KCL_ERROR_MSGS[self.lang]

    @PreCheck((lambda v: CheckRules.check_locale(v)), "lang")
    def switch_lang(self, lang: str):
        try:
            self.KCL_ERROR_MSG = self.ALL_KCL_ERROR_MSGS[lang]
        except KeyError:
            print(
                f"KCLVM does not support '{lang}', "
                f"KCLVM have automatically switched to {ErrMsgLang.EN_US_UTF_8}",
                file=sys.stderr,
            )
            self.lang = ErrMsgLang.EN_US_UTF_8
            self.KCL_ERROR_MSG = self.ALL_KCL_ERROR_MSGS[self.lang]

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, ErrMsg)), "err_msg")
    @PreCheck((lambda v: CheckRules.check_locale(v)), "lang")
    def append_lang(self, lang: str, err_msg: ErrMsg):
        if lang in self.ALL_KCL_ERROR_MSGS.keys():
            print(
                f"KCLVM currently supports {lang}, "
                f"If you want to change the language pack, "
                f"please use the method update_lang",
                file=sys.stderr,
            )
            return
        else:
            self.ALL_KCL_ERROR_MSGS[lang] = err_msg

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, ErrMsg)), "err_msg")
    @PreCheck((lambda v: CheckRules.check_locale(v)), "lang")
    def update_lang(self, lang: str, err_msg: ErrMsg):
        self.ALL_KCL_ERROR_MSGS[lang] = err_msg

    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    @PostCheck(lambda result: result == ErrEwcode.KCLException_Ew)
    def get_defaule_ewcode(self) -> str:
        return self.KCL_ERROR_MSG.get_defaule_ewcode()

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    @PostSimpleExprCheck(
        (lambda inputs, result: result == inputs["err_id"][:1] + "0000"), ["err_id"]
    )
    def get_err_cate_ewcode(self, err_id: str) -> str:
        return self.KCL_ERROR_MSG.get_err_cate_ewcode(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    @PostSimpleExprCheck(
        (lambda inputs, result: result == "0" + inputs["err_id"][1:2] + "000"),
        ["err_id"],
    )
    def get_err_type_ewcode(self, err_id: str) -> str:
        return self.KCL_ERROR_MSG.get_err_type_ewcode(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    def get_err_cate_msg_by_errid(self, err_id: str) -> str:
        return self.KCL_ERROR_MSG.get_err_cate_msg_by_errid(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    def get_err_type_msg_by_errid(self, err_id: str) -> str:
        return self.KCL_ERROR_MSG.get_err_type_msg_by_errid(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    def get_err_msg_by_errid(self, err_id: str) -> str:
        return self.KCL_ERROR_MSG.get_err_msg_by_errid(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    def get_defaule_arg_msg_by_errid(self, err_id: str) -> str:
        return self.KCL_ERROR_MSG.get_defaule_arg_msg_by_errid(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    @PostSimpleExprCheck(
        (lambda inputs, result: result == "[" + inputs["err_id"] + "]"), ["err_id"]
    )
    def get_err_code_by_errid(self, err_id: str) -> str:
        return self.KCL_ERROR_MSG.get_err_code_by_errid(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, bool))
    def is_warning(self, err_id: str) -> bool:
        return self.KCL_ERROR_MSG.is_warning(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, bool))
    def is_syntax_error(self, err_id: str) -> bool:
        return self.KCL_ERROR_MSG.is_syntax_error(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, bool))
    def is_compiler_error(self, err_id: str) -> bool:
        return self.KCL_ERROR_MSG.is_compiler_error(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, bool))
    def is_runtime_error(self, err_id: str) -> bool:
        return self.KCL_ERROR_MSG.is_runtime_error(err_id)

    @PreCheck((lambda v: ErrEwcode.contains(v)), "err_id")
    @PostCheck(lambda result: CheckRules.check_type_not_none(result, str))
    def get_err_name_by_ewcode(self, err_id: str):
        return self.KCL_ERROR_MSG.get_err_name_by_ewcode(err_id)


KCL_ERR_MSG = KCLErrMsgManager()



# Copyright 2020 The KCL Authors. All rights reserved.

from typing import List, Dict, Optional, cast
from dataclasses import dataclass, field

import kclvm.kcl.ast as ast
import kclvm.api.object as objpkg
import kclvm.api.object.internal as internal
import kclvm.compiler.extension.builtin.builtin as builtin
import kclvm.compiler.extension.plugin as plugin

from .type import (
    Type,
    ANY_TYPE,
    STR_TYPE,
    BOOL_TYPE,
    INT_TYPE,
    FLOAT_TYPE,
    LIST_ANY_TYPE,
    LIST_STR_TYPE,
    DICT_ANY_ANY_TYPE,
    ITERABLE_TYPE_STR,
    ITERABLE_TYPE,
    NUMBER_TYPE_STR,
    NUMBER_TYPE,
)


@dataclass
class ScopeObject:
    """The object stored in the scope

    Parameters
    ----------
    name: str
        The scope object name.
    node: ast.AST
        The scope object AST node reference.
    type: Type
        The type of the scope object.
    pos: ast.Position
        The scope object start position.
    end: ast.Position
        The scope object end position.
    """

    name: str
    node: Optional[ast.AST]
    type: Type
    pos: ast.Position = None
    end: ast.Position = None

    def check_pos_valid(self):
        return (
            self.node
            and self.node.pos
            and self.node.pos.is_valid()
            and self.node.end_pos
            and self.node.end_pos.is_valid()
        )


@dataclass
class Scope:
    """A Scope maintains a set of objects and links to its containing
    (parent) and contained (children) scopes. Objects may be inserted
    and looked up by name. The zero value for Scope is a ready-to-use
    empty scope.

    Parameters
    ----------
    parent: Scope
        The parent scope.
    node: ast.AST
        The scope AST node reference.
    children: Type:
        The child scope list.
    elems: Dict[str, ScopeObject]
        The scope object mapping with its name.
    pos: ast.Position
        The scope start position.
    end: ast.Position
        The scope end position.
    """

    parent: "Scope" = None
    node: ast.AST = None
    children: List["Scope"] = field(default_factory=list)
    elems: Dict[str, ScopeObject] = field(default_factory=dict)
    pos: ast.Position = None
    end: ast.Position = None

    def contains_pos(self, pos: ast.Position) -> bool:
        """
        check if current scope contains a position
        :param pos: the given position
        :return: if current scope contains the given position
        """
        if isinstance(self.node, ast.SchemaStmt):
            for item in [
                *(self.node.body if self.node.body else []),
                *(self.node.checks if self.node.checks else []),
                self.node.index_signature,
            ]:
                if item and item.contains_pos(pos):
                    return True
            return False
        elif isinstance(self.node, ast.RuleStmt):
            for item in self.node.checks or []:
                if item and item.contains_pos(pos):
                    return True
            return False
        return self.pos and self.pos.less_equal(pos) and pos.less_equal(self.end)

    def inner_most(self, pos: ast.Position) -> Optional["Scope"]:
        # self is BUILTIN_SCOPE
        if self.parent is None:
            for child in self.children or []:
                if child.contains_pos(pos):
                    return child.inner_most(pos)
            return None
        # self is not BUILTIN_SCOPE
        if self.contains_pos(pos):
            for child in self.children or []:
                if child.contains_pos(pos):
                    return child.inner_most(pos)
            return self
        return None

    def get_enclosing_scope(self) -> Optional["Scope"]:
        return self.parent

    def get_parent_schema_scope(
        self, program_scope: "ProgramScope"
    ) -> Optional["Scope"]:
        if (
            self.node
            and isinstance(self.node, ast.SchemaStmt)
            and self.node.parent_name
        ):
            schema = self.parent.elems[self.node.name]
            if not isinstance(schema.type, objpkg.KCLSchemaDefTypeObject):
                return None
            schema_type_def_obj = cast(objpkg.KCLSchemaDefTypeObject, schema.type)
            if (
                not schema_type_def_obj.schema_type
                or not schema_type_def_obj.schema_type.base
            ):
                return None
            base_type_obj = schema_type_def_obj.schema_type.base
            # the schema and its base schema are in the same scope
            if schema_type_def_obj.schema_type.pkgpath == base_type_obj.pkgpath:
                return self.parent.search_child_scope_by_name(base_type_obj.name)
            # the schema and its base schema are in the different scopes
            base_pkg_scope = program_scope.scope_map.get(base_type_obj.pkgpath)
            if not base_pkg_scope:
                return None
            return base_pkg_scope.search_child_scope_by_name(base_type_obj.name)
        return None

    def search_child_scope_by_name(self, name: str) -> Optional["Scope"]:
        assert name
        if name not in self.elems:
            return None
        elem: ScopeObject = self.elems[name]
        for child in self.children:
            if child.node == elem.node:
                return child
        return None


@dataclass
class PackageScope(Scope):
    file_begin_position_map: Dict[str, ast.Position] = field(default_factory=dict)
    file_end_position_map: Dict[str, ast.Position] = field(default_factory=dict)

    def contains_pos(self, pos: ast.Position) -> bool:
        """
        check if current package scope contains a position
        :param pos: the given position
        :return: if current package scope contains the given position
        """
        assert pos.filename
        file_begin_pos = self.file_begin_position_map.get(pos.filename)
        file_end_pos = self.file_end_position_map.get(pos.filename)
        return (
            file_begin_pos
            and file_end_pos
            and pos
            and file_begin_pos.less_equal(pos)
            and pos.less_equal(file_end_pos)
        )


# Decorator scope including decorator function type such as `deprecated`.
DECORATOR_SCOPE: Scope = Scope(
    elems={
        internal.Deprecated.NAME: ScopeObject(
            name=internal.Deprecated.NAME,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name=internal.Deprecated.NAME,
                params=[
                    objpkg.Parameter(
                        name="version",
                        type_annotation="str",
                        type=STR_TYPE,
                    ),
                    objpkg.Parameter(
                        name="reason",
                        type_annotation="str",
                        type=STR_TYPE,
                    ),
                    objpkg.Parameter(
                        name="strict",
                        type_annotation="bool",
                        type=BOOL_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=ANY_TYPE,
                doc=internal.Deprecated.__doc__,
            ),
        ),
        internal.Info.NAME: ScopeObject(
            name=internal.Info.NAME,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name=internal.Info.NAME,
                params=[],
                self_type=ANY_TYPE,
                return_type=ANY_TYPE,
                doc=internal.Info.__doc__,
            ),
        ),
    }
)


# Builtin-function types table
BUILTIN_SCOPE: Scope = Scope(
    elems={
        # option(key: str, *, type: str = "", required: bool = False, default: any = None, help: str = "", file: str = "", line: int = 0) -> any
        "option": ScopeObject(
            name="option",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="option",
                params=[
                    objpkg.Parameter(
                        name="key",
                        value=None,
                        type_annotation="str",
                        type=STR_TYPE,
                    ),
                    objpkg.Parameter(
                        name="type",
                        value=objpkg.KCLStringObject(""),
                        type_annotation="str",
                        type=STR_TYPE,
                    ),
                    objpkg.Parameter(
                        name="required",
                        value=objpkg.KCLFalseObject.instance(),
                        type_annotation="bool",
                        type=BOOL_TYPE,
                    ),
                    objpkg.Parameter(
                        name="default",
                        value=objpkg.KCLNoneObject.instance(),
                        type_annotation="any",
                        type=ANY_TYPE,
                    ),
                    objpkg.Parameter(
                        name="help",
                        value=objpkg.KCLStringObject(""),
                        type_annotation="str",
                        type=STR_TYPE,
                    ),
                    objpkg.Parameter(
                        name="help",
                        value=objpkg.KCLStringObject(""),
                        type_annotation="str",
                        type=STR_TYPE,
                    ),
                    objpkg.Parameter(
                        name="file",
                        value=objpkg.KCLStringObject(""),
                        type_annotation="str",
                        type=STR_TYPE,
                    ),
                    objpkg.Parameter(
                        name="line",
                        value=objpkg.KCLIntObject(0),
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=ANY_TYPE,
                doc=builtin.KMANGLED_option.__doc__,
            ),
        ),
        # print(value, ..., sep=' ', end='\n') -> NONE_TYPE
        "print": ScopeObject(
            name="print",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="print",
                params=[],
                self_type=ANY_TYPE,
                return_type=ANY_TYPE,
                doc=builtin.KMANGLED_print.__doc__,
                is_variadic=True,
            ),
        ),
        # multiplyof(a: int, b: int) -> bool
        "multiplyof": ScopeObject(
            name="multiplyof",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="multiplyof",
                params=[
                    objpkg.Parameter(
                        name="a",
                        value=None,
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                    objpkg.Parameter(
                        name="b",
                        value=None,
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=INT_TYPE,
                doc=builtin.KMANGLED_multiplyof.__doc__,
            ),
        ),
        # isunique(inval: List[Any]) -> bool
        "isunique": ScopeObject(
            name="isunique",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="isunique",
                params=[
                    objpkg.Parameter(
                        name="inval",
                        value=None,
                        type_annotation="[]",
                        type=LIST_ANY_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=BOOL_TYPE,
                doc=builtin.KMANGLED_isunique.__doc__,
            ),
        ),
        # len(inval: Union[dict, list, schema, str]) -> int
        "len": ScopeObject(
            name="len",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="len",
                params=[
                    objpkg.Parameter(
                        name="inval",
                        value=None,
                        type_annotation=ITERABLE_TYPE_STR,
                        type=ITERABLE_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=INT_TYPE,
                doc=builtin.KMANGLED_len.__doc__,
            ),
        ),
        # abs(inval: Union[int, float, bool]) -> Union[int, float, bool]
        "abs": ScopeObject(
            name="abs",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="abs",
                params=[
                    objpkg.Parameter(
                        name="inval",
                        value=None,
                        type_annotation="any",
                        type=ANY_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=ANY_TYPE,
                doc=builtin.KMANGLED_abs.__doc__,
            ),
        ),
        # all_true(inval: Union[dict, list, schema, str]) -> bool
        "all_true": ScopeObject(
            name="all_true",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="all_true",
                params=[
                    objpkg.Parameter(
                        name="inval",
                        value=None,
                        type_annotation="[]",
                        type=LIST_ANY_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=BOOL_TYPE,
                doc=builtin.KMANGLED_all_true.__doc__,
            ),
        ),
        # any_true(inval: Union[dict, list, schema, str]) -> bool
        "any_true": ScopeObject(
            name="any_true",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="any_true",
                params=[
                    objpkg.Parameter(
                        name="inval",
                        value=None,
                        type_annotation="[]",
                        type=LIST_ANY_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=BOOL_TYPE,
                doc=builtin.KMANGLED_any_true.__doc__,
            ),
        ),
        # hex(number: int) -> str
        "hex": ScopeObject(
            name="hex",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="hex",
                params=[
                    objpkg.Parameter(
                        name="number",
                        value=None,
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=STR_TYPE,
                doc=builtin.KMANGLED_hex.__doc__,
            ),
        ),
        # bin(number: int) -> str
        "bin": ScopeObject(
            name="bin",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="bin",
                params=[
                    objpkg.Parameter(
                        name="number",
                        value=None,
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=STR_TYPE,
                doc=builtin.KMANGLED_bin.__doc__,
            ),
        ),
        # oct(number: int) -> str
        "oct": ScopeObject(
            name="oct",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="oct",
                params=[
                    objpkg.Parameter(
                        name="number",
                        value=None,
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=STR_TYPE,
                doc=builtin.KMANGLED_oct.__doc__,
            ),
        ),
        # ord(c: str) -> int
        "ord": ScopeObject(
            name="ord",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="ord",
                params=[
                    objpkg.Parameter(
                        name="number",
                        value=None,
                        type_annotation="int",
                        type=STR_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=INT_TYPE,
                doc=builtin.KMANGLED_ord.__doc__,
            ),
        ),
        # sorted(inval: Union[dict, list, schema, str]) -> List[Any]
        "sorted": ScopeObject(
            name="sorted",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="sorted",
                params=[
                    objpkg.Parameter(
                        name="inval",
                        value=None,
                        type_annotation=ITERABLE_TYPE_STR,
                        type=ITERABLE_TYPE,
                    ),
                    objpkg.Parameter(
                        name="reverse",
                        value=objpkg.KCLFalseObject.instance(),
                        type_annotation="bool",
                        type=BOOL_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=LIST_ANY_TYPE,
                doc=builtin.KMANGLED_sorted.__doc__,
            ),
        ),
        # range(start: int, stop: int, step: int = None) -> list
        "range": ScopeObject(
            name="range",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="range",
                params=[
                    objpkg.Parameter(
                        name="start",
                        value=None,
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                    objpkg.Parameter(
                        name="stop",
                        value=None,
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                    objpkg.Parameter(
                        name="step",
                        value=objpkg.KCLNoneObject.instance(),
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=objpkg.KCLListTypeObject(item_type=INT_TYPE),
                doc=builtin.KMANGLED_range.__doc__,
            ),
        ),
        # max(iterable: List[Any]) -> Any
        "max": ScopeObject(
            name="max",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="max",
                params=[],
                self_type=ANY_TYPE,
                return_type=ANY_TYPE,
                doc=builtin.KMANGLED_max.__doc__,
            ),
        ),
        # min(iterable: List[Any]) -> Any
        "min": ScopeObject(
            name="min",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="min",
                params=[],
                self_type=ANY_TYPE,
                return_type=ANY_TYPE,
                doc=builtin.KMANGLED_min.__doc__,
            ),
        ),
        # sum(iterable: List[Any], start: Any = 0) -> Any
        "sum": ScopeObject(
            name="sum",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="sum",
                params=[
                    objpkg.Parameter(
                        name="iterable",
                        value=None,
                        type_annotation="[]",
                        type=LIST_ANY_TYPE,
                    ),
                    objpkg.Parameter(
                        name="start",
                        value=objpkg.KCLIntObject(0),
                        type_annotation="any",
                        type=ANY_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=ANY_TYPE,
                doc=builtin.KMANGLED_sum.__doc__,
            ),
        ),
        # pow(x: number, y: number, z: number) -> number
        "pow": ScopeObject(
            name="pow",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="pow",
                params=[
                    objpkg.Parameter(
                        name="x",
                        value=None,
                        type_annotation=NUMBER_TYPE_STR,
                        type=NUMBER_TYPE,
                    ),
                    objpkg.Parameter(
                        name="y",
                        value=None,
                        type_annotation=NUMBER_TYPE_STR,
                        type=NUMBER_TYPE,
                    ),
                    objpkg.Parameter(
                        name="z",
                        value=None,
                        type_annotation=NUMBER_TYPE_STR,
                        type=NUMBER_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=NUMBER_TYPE,
                doc=builtin.KMANGLED_pow.__doc__,
            ),
        ),
        # round(number: number, ndigits: int) -> number
        "round": ScopeObject(
            name="round",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="round",
                params=[
                    objpkg.Parameter(
                        name="number",
                        value=None,
                        type_annotation=NUMBER_TYPE_STR,
                        type=NUMBER_TYPE,
                    ),
                    objpkg.Parameter(
                        name="ndigits",
                        value=objpkg.KCLNoneObject.instance(),
                        type_annotation="int",
                        type=INT_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=NUMBER_TYPE,
                doc=builtin.KMANGLED_round.__doc__,
            ),
        ),
        # zip(*args) -> List[Any]
        "zip": ScopeObject(
            name="zip",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="zip",
                params=[],
                self_type=ANY_TYPE,
                return_type=LIST_ANY_TYPE,
                doc=builtin.KMANGLED_zip.__doc__,
                is_variadic=True,
            ),
        ),
        # int(number: number|str) -> int
        "int": ScopeObject(
            name="int",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="int",
                params=[
                    objpkg.Parameter(
                        name="number",
                        value=None,
                        type_annotation="any",
                        type=ANY_TYPE,
                    ),
                    objpkg.Parameter(
                        name="base",
                        value=objpkg.KCLIntObject(10),
                        type_annotation="int",
                        type=ANY_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=INT_TYPE,
                doc=builtin.KMANGLED_int.__doc__,
            ),
        ),
        # float(number: number|str) -> float
        "float": ScopeObject(
            name="float",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="float",
                params=[
                    objpkg.Parameter(
                        name="number",
                        value=None,
                        type_annotation="any",
                        type=ANY_TYPE,
                    )
                ],
                self_type=ANY_TYPE,
                return_type=FLOAT_TYPE,
                doc=builtin.KMANGLED_float.__doc__,
            ),
        ),
        # list(x: any) -> []
        "list": ScopeObject(
            name="list",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="list",
                params=[
                    objpkg.Parameter(
                        name="x",
                        value=None,
                        type_annotation="any",
                        type=ANY_TYPE,
                    )
                ],
                self_type=ANY_TYPE,
                return_type=LIST_ANY_TYPE,
                doc=builtin.KMANGLED_list.__doc__,
            ),
        ),
        # dict(x: any) -> {:}
        "dict": ScopeObject(
            name="dict",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="dict",
                params=[
                    objpkg.Parameter(
                        name="x",
                        value=None,
                        type_annotation="any",
                        type=ANY_TYPE,
                    )
                ],
                self_type=ANY_TYPE,
                return_type=DICT_ANY_ANY_TYPE,
                doc=builtin.KMANGLED_dict.__doc__,
            ),
        ),
        # bool(x: any) -> bool
        "bool": ScopeObject(
            name="bool",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="bool",
                params=[
                    objpkg.Parameter(
                        name="x",
                        value=None,
                        type_annotation="any",
                        type=ANY_TYPE,
                    )
                ],
                self_type=ANY_TYPE,
                return_type=BOOL_TYPE,
                doc=builtin.KMANGLED_bool.__doc__,
            ),
        ),
        # str(obj: any) -> str
        "str": ScopeObject(
            name="str",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="str",
                params=[
                    objpkg.Parameter(
                        name="x",
                        value=None,
                        type_annotation="any",
                        type=ANY_TYPE,
                    )
                ],
                self_type=ANY_TYPE,
                return_type=STR_TYPE,
                doc=builtin.KMANGLED_str.__doc__,
            ),
        ),
        # typeof(x: any, *, full_name: bool = False) -> str:
        "typeof": ScopeObject(
            name="typeof",
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="typeof",
                params=[
                    objpkg.Parameter(
                        name="x",
                        value=None,
                        type_annotation="any",
                        type=ANY_TYPE,
                    ),
                    objpkg.Parameter(
                        name="full_name",
                        value=objpkg.KCLFalseObject.instance(),
                        type_annotation="bool",
                        type=BOOL_TYPE,
                    ),
                ],
                self_type=ANY_TYPE,
                return_type=STR_TYPE,
                doc=builtin.KMANGLED_typeof.__doc__,
            ),
        ),
    }
)

# System module types table
MODULE_SCOPE_MAPPING: Dict[str, Scope] = {
    **{name: Scope(elems={}) for name in builtin.STANDARD_SYSTEM_MODULES},
    **{
        "units": Scope(
            elems={
                "NumberMultiplier": ScopeObject(
                    name="NumberMultiplier",
                    node=None,
                    type=objpkg.KCLNumberMultiplierTypeObject(),
                ),
            }
        ),
    },
}

try:
    # Plugin module types table
    PLUGIN_SCOPE_MAPPING: Dict[str, Scope] = {
        "{}".format(plugin.PLUGIN_MODULE_NAME + name): Scope(elems={})
        for name in plugin.get_plugin_names()
    }
except Exception:
    PLUGIN_SCOPE_MAPPING: Dict[str, Scope] = {}


# Member function or property types table e.g., str.replace and schema.instances
SCHEMA_TYPE_MEMBER_SCOPE: Scope = Scope(
    elems={
        "instances": ScopeObject(
            name=objpkg.KCLSchemaTypeObject.instances.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name=objpkg.KCLSchemaTypeObject.instances.__name__,
                params=[],
                self_type=DICT_ANY_ANY_TYPE,
                return_type=LIST_ANY_TYPE,
                doc=objpkg.KCLSchemaTypeObject.instances.__doc__,
            ),
        )
    }
)

STR_TYPE_MEMBER_SCOPE: Scope = Scope(
    elems={
        "capitalize": ScopeObject(
            name="".capitalize.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".capitalize.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".capitalize.__doc__,
            ),
        ),
        "count": ScopeObject(
            name="".count.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".count.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=INT_TYPE,
                doc="".count.__doc__,
            ),
        ),
        "endswith": ScopeObject(
            name="".endswith.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".endswith.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=BOOL_TYPE,
                doc="".endswith.__doc__,
            ),
        ),
        "find": ScopeObject(
            name="".find.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".find.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=INT_TYPE,
                doc="".find.__doc__,
            ),
        ),
        "format": ScopeObject(
            name="".format.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".format.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".format.__doc__,
            ),
        ),
        "index": ScopeObject(
            name="".index.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".index.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=INT_TYPE,
                doc="".index.__doc__,
            ),
        ),
        "isalnum": ScopeObject(
            name="".isalnum.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".isalnum.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=BOOL_TYPE,
                doc="".isalnum.__doc__,
            ),
        ),
        "isalpha": ScopeObject(
            name="".isalpha.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".isalpha.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=BOOL_TYPE,
                doc="".isalpha.__doc__,
            ),
        ),
        "isdigit": ScopeObject(
            name="".isdigit.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".isdigit.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=BOOL_TYPE,
                doc="".isdigit.__doc__,
            ),
        ),
        "islower": ScopeObject(
            name="".islower.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".islower.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=BOOL_TYPE,
                doc="".islower.__doc__,
            ),
        ),
        "isspace": ScopeObject(
            name="".isspace.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".isspace.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=BOOL_TYPE,
                doc="".isspace.__doc__,
            ),
        ),
        "istitle": ScopeObject(
            name="".istitle.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".istitle.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=BOOL_TYPE,
                doc="".istitle.__doc__,
            ),
        ),
        "isupper": ScopeObject(
            name="".isupper.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".isupper.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=BOOL_TYPE,
                doc="".isupper.__doc__,
            ),
        ),
        "join": ScopeObject(
            name="".join.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".join.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".join.__doc__,
            ),
        ),
        "lower": ScopeObject(
            name="".lower.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".lower.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".lower.__doc__,
            ),
        ),
        "upper": ScopeObject(
            name="".upper.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".upper.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".upper.__doc__,
            ),
        ),
        "lstrip": ScopeObject(
            name="".lstrip.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".lstrip.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".lstrip.__doc__,
            ),
        ),
        "rstrip": ScopeObject(
            name="".rstrip.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".rstrip.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".rstrip.__doc__,
            ),
        ),
        "replace": ScopeObject(
            name="".replace.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".replace.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".replace.__doc__,
            ),
        ),
        "rfind": ScopeObject(
            name="".rfind.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".rfind.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=INT_TYPE,
                doc="".rfind.__doc__,
            ),
        ),
        "rindex": ScopeObject(
            name="".rindex.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".rindex.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=INT_TYPE,
                doc="".rindex.__doc__,
            ),
        ),
        "rsplit": ScopeObject(
            name="".rsplit.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".rsplit.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=LIST_STR_TYPE,
                doc="".rsplit.__doc__,
            ),
        ),
        "split": ScopeObject(
            name="".split.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".split.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=LIST_STR_TYPE,
                doc="".split.__doc__,
            ),
        ),
        "splitlines": ScopeObject(
            name="".splitlines.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".splitlines.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=LIST_STR_TYPE,
                doc="".splitlines.__doc__,
            ),
        ),
        "startswith": ScopeObject(
            name="".startswith.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".startswith.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=BOOL_TYPE,
                doc="".startswith.__doc__,
            ),
        ),
        "strip": ScopeObject(
            name="".strip.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".strip.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".strip.__doc__,
            ),
        ),
        "title": ScopeObject(
            name="".title.__name__,
            node=None,
            type=objpkg.KCLFunctionTypeObject(
                name="".title.__name__,
                params=[],
                self_type=STR_TYPE,
                return_type=STR_TYPE,
                doc="".title.__doc__,
            ),
        ),
    }
)


@dataclass
class ProgramScope:
    scope_map: Dict[str, Scope]
    builtin_scope: Scope = field(default_factory=lambda: BUILTIN_SCOPE)
    plugin_scope_map: Scope = field(default_factory=lambda: PLUGIN_SCOPE_MAPPING)
    system_module_scope_map: Dict[str, Scope] = field(
        default_factory=lambda: MODULE_SCOPE_MAPPING
    )
    member_scope_map: Dict[str, Scope] = field(
        default_factory=lambda: {
            objpkg.KCLTypeKind.StrKind: SCHEMA_TYPE_MEMBER_SCOPE,
            objpkg.KCLTypeKind.SchemaKind: STR_TYPE_MEMBER_SCOPE,
        }
    )
    schema_reference: objpkg.SchemaTypeRefGraph = field(default_factory=lambda: None)

    @property
    def main_scope(self) -> Scope:
        return self.scope_map[ast.Program.MAIN_PKGPATH]

    @property
    def pkgpaths(self) -> List[str]:
        return list(self.scope_map.keys())



from io import StringIO
from pathlib import Path
import ruamel.yaml as yaml

from .util import filter_fields

_yaml = yaml.YAML()
_yaml.representer.add_representer(
    str,
    lambda dumper, data: dumper.represent_scalar(
        "tag:yaml.org,2002:str", data, style="|"
    )
    if "\n" in data
    else dumper.represent_str(data),
)
# Convert None to null
_yaml.representer.add_representer(
    type(None),
    lambda dumper, data: dumper.represent_scalar("tag:yaml.org,2002:null", "null"),
)


def KMANGLED_encode(data, sort_keys=False, ignore_private=False, ignore_none=False):
    buffer = StringIO()
    data_filtered = filter_fields(data, ignore_private, ignore_none)
    if sort_keys:
        sorted_dict = yaml.comments.CommentedMap()
        for k in sorted(data_filtered):
            sorted_dict[k] = data_filtered[k]
        _yaml.dump(sorted_dict, buffer)
    else:
        _yaml.dump(data_filtered, buffer)
    return buffer.getvalue()


def KMANGLED_decode(value: str):
    buffer = StringIO(value)
    data = yaml.safe_load(buffer)
    return data


def KMANGLED_dump_to_file(
    data, filename: str, sort_keys=False, ignore_private=False, ignore_none=False
):
    yaml_str = KMANGLED_encode(data, sort_keys, ignore_private, ignore_none)
    Path(filename).write_text(yaml_str)


def KMANGLED_load_all(value: str):
    buffer = StringIO(value)
    data = yaml.safe_load_all(buffer)
    return data



from copy import deepcopy

import kclvm.api.object.internal.common as common


def KMANGLED_union_all(data: list) -> dict:
    if not data:
        return {}
    data_copy = deepcopy(data)
    value = data[0]
    for d in data_copy[1:]:
        common.union(value, d)
    return value



# Copyright 2021 The KCL Authors. All rights reserved.

import pathlib
from typing import List

import kclvm.config
import kclvm.internal.util as util
import kcl_lib.api.spec_pb2 as pb2


KCL_MOD_PATH_ENV = "${KCL_MOD}"


def load_settings_files(
    work_dir: str, files: List[str]
) -> pb2.LoadSettingsFiles_Result:
    """Load KCL CLI config from the setting files.

    Parameter
    ---------
    work_dir : str
        The kcl run work directory.
    files:
        The setting YAML files.

    Returns
    -------
    result: LoadSettingsFiles_Result
        The merged kcl singleton config.
    """
    from kclvm.compiler.vfs import GetPkgRoot

    if not files:
        return pb2.LoadSettingsFiles_Result(
            kcl_cli_configs=pb2.CliConfig(), kcl_options=[]
        )
    key_value_pairs = [
        pb2.KeyValuePair(key=k, value=v)
        for k, v in util.merge_option_same_keys(
            kclvm.config.KCLCLISettingAction().deal(files)[0]
        ).items()
    ]
    if work_dir or kclvm.config.current_path:
        files = [
            str(
                pathlib.Path(work_dir)
                .joinpath(
                    str(x).replace(
                        KCL_MOD_PATH_ENV,
                        GetPkgRoot(work_dir or kclvm.config.current_path or files[0])
                        or "",
                    )
                )
                .resolve()
            )
            for x in kclvm.config.input_file
        ]
    return pb2.LoadSettingsFiles_Result(
        kcl_cli_configs=pb2.CliConfig(
            files=files,
            output=kclvm.config.output,
            overrides=kclvm.config.overrides,
            path_selector=kclvm.config.path_selector,
            strict_range_check=kclvm.config.strict_range_check,
            disable_none=kclvm.config.disable_none,
            verbose=kclvm.config.verbose,
            debug=kclvm.config.debug,
        ),
        kcl_options=key_value_pairs,
    )



import sys
from typing import List

from kclvm.tools.lint.reporters.base_reporter import BaseReporter
from kclvm.tools.lint.message.message import Message

LINT_THEME: dict = {
    "ID": "\033[0;92m{}\033[0m",  # green
    "ERROR": "\033[0;91m{}\033[0m",  # red
    "WARNING": "\033[0;93m{}\033[0m",  # yellow
    "FILE_NAME": "\033[0;94m{}\033[0m",  # blue
    "LINE_COLUMN": "\033[0;96m{}\033[0m",  # cyan
    "MARK": "\033[0;31m{}\033[0m",  # red
    "NUMBER": "\033[0;31m{}\033[0m",  # red
}


def color(content: str, content_type: str):
    return LINT_THEME[content_type].format(content)


def msg_with_color(msg: Message):
    return (
        color(msg.file, "FILE_NAME")
        + ":"
        + color(msg.pos[0], "LINE_COLUMN")
        + ":"
        + color(msg.pos[1], "LINE_COLUMN")
        + ": "
        + color(msg.msg_id, "ID")
        + " "
        + msg.msg
        + "\n"
        + msg.source_code
        + "\n"
        + (msg.pos[1] - 1) * " "
        + color("^", "MARK")
    )


class STDOUTReporter(BaseReporter):
    def __init__(self, linter, output=None, encoding=None):
        self.name = "stdout_reporter"
        super().__init__(linter, output, encoding)

    def print_msg(self, msgs: List[Message], file=sys.stdout):
        """
        Print msgs with color.Because CI cannot parse color information, e.g. [0;31m{,
        it is not enabled temporarily

        for msg in msgs:
            print((msg_with_color(msg) if file.isatty() else str(msg)) + "\n")
        print(
            "Check total "
            + (
                color(len(self.linter.file_list), "NUMBER")
                if file.isatty()
                else str(len(self.linter.file_list))
            )
            + " files:"
        )
        for k, v in self.linter.msgs_map.items():
            print(("{:<19}".format(color(v, "NUMBER"))
                   + color(k, "ID")
                   + ": "
                   + self.linter.MSGS[k][1]
                   ) if file.isatty() else ("{:<8}{}: {}".format(v, k, self.linter.MSGS[k][1])))
        print("KCL Lint: "
              + (color(len(self.linter.msgs), "NUMBER") if file.isatty() else str(len(self.linter.msgs)))
              + " problems")
        """
        for msg in msgs:
            print(msg)
            print()
        print(f"Check total {len(self.linter.file_list)} files:")
        for k, v in self.linter.msgs_map.items():
            print("{:<8}{}: {}".format(v, k, self.linter.MSGS[k][1]))
        print(f"KCL Lint: {len(self.linter.msgs)} problems")



from typing import List, cast, Optional
from pygls.lsp.types.basic_structures import Location

import kclvm.kcl.ast as ast
import kclvm.kcl.types.scope as scope
import kclvm.tools.langserver.common as common
from kclvm.api.object.object import KCLTypeKind, KCLModuleTypeObject
from kclvm.api.object.schema import KCLSchemaTypeObject, KCLSchemaDefTypeObject


def definition(
    pos: ast.Position, code: str = None
) -> (Optional[ast.AST], Optional[scope.ScopeObject]):
    prog: ast.Program
    prog, leaf_node = common.pos_to_node(pos, code)
    if not leaf_node:
        # no name node at the position
        return None, None
    parent: ast.AST = leaf_node.parent
    if isinstance(leaf_node, ast.Name):
        if (
            parent.type == "Identifier"
            and parent.parent.type == "ConfigEntry"
            and parent is parent.parent.key
        ):
            identifier: ast.Identifier = cast(ast.Identifier, parent)
            _, prog_scope = common.file_or_prog_to_scope(prog)
            schema_expr: ast.SchemaExpr = leaf_node.find_nearest_parent_by_type(
                ast.SchemaExpr
            )
            if schema_expr:
                schema_name: ast.Identifier = schema_expr.name
                schema_scope_obj = find_declaration(
                    schema_name, schema_name.name_nodes[-1], prog_scope
                )
                top_attr = find_inner_name(
                    schema_scope_obj, identifier.name_nodes[0], prog_scope
                )
                result_obj = find_declaration_by_scope_obj(
                    identifier=identifier,
                    name_node=leaf_node,
                    top_name_obj=top_attr,
                    prog_scope=prog_scope,
                )
                return leaf_node, result_obj
        if parent.type == "Identifier" and (
            parent.parent.type != "ConfigEntry" or parent is parent.parent.value
        ):
            identifier: ast.Identifier = cast(ast.Identifier, parent)
            _, prog_scope = common.file_or_prog_to_scope(prog)
            declaration = find_declaration(identifier, leaf_node, prog_scope)
            return leaf_node, declaration
    return leaf_node, None


def go_to_def(pos: ast.Position, code: str = None) -> List[Location]:
    prog: ast.Program
    prog, leaf_node = common.pos_to_node(pos, code)
    if not leaf_node:
        # no name node at the position
        return []
    parent: ast.AST = leaf_node.parent
    if isinstance(leaf_node, ast.Name):
        if parent.type == "ImportStmt":
            import_stmt: ast.ImportStmt = cast(ast.ImportStmt, parent)
            if leaf_node in import_stmt.path_nodes:
                index = import_stmt.path_nodes.index(leaf_node)
                if index == len(import_stmt.path_nodes) - 1:
                    # this might be a module name, return the target module file path
                    loc = common.pkgpath_to_location(
                        root=prog.root, pkgpath=import_stmt.path
                    )
                    return [loc] if loc else []
            return [common.node_to_location(leaf_node)]
        if (
            parent.type == "Identifier"
            and parent.parent.type == "ConfigEntry"
            and parent is parent.parent.key
        ):
            identifier: ast.Identifier = cast(ast.Identifier, parent)
            _, prog_scope = common.file_or_prog_to_scope(prog)
            schema_expr: ast.SchemaExpr = leaf_node.find_nearest_parent_by_type(
                ast.SchemaExpr
            )
            if schema_expr:
                schema_name: ast.Identifier = schema_expr.name
                schema_scope_obj = find_declaration(
                    schema_name, schema_name.name_nodes[-1], prog_scope
                )
                top_attr = find_inner_name(
                    schema_scope_obj, identifier.name_nodes[0], prog_scope
                )
                result_obj = find_declaration_by_scope_obj(
                    identifier=identifier,
                    name_node=leaf_node,
                    top_name_obj=top_attr,
                    prog_scope=prog_scope,
                )
                loc = common.scope_obj_to_location(result_obj)
                return [loc] if loc else []
        if parent.type == "Identifier" and (
            parent.parent.type != "ConfigEntry" or parent is parent.parent.value
        ):
            identifier: ast.Identifier = cast(ast.Identifier, parent)
            _, prog_scope = common.file_or_prog_to_scope(prog)
            declaration = find_declaration(identifier, leaf_node, prog_scope)
            loc = common.scope_obj_to_location(declaration)
            return [loc] if loc else []
    return [common.node_to_location(leaf_node)]


def find_declaration(
    identifier: ast.Identifier, name_node: ast.Name, prog_scope: scope.ProgramScope
) -> Optional[scope.ScopeObject]:
    if not identifier or not name_node or not prog_scope:
        return None
    top_name = identifier.name_nodes[0]
    top_name_obj = find_declaration_obj_by_pos_and_name(
        top_name.pos, top_name.value, prog_scope
    )
    return find_declaration_by_scope_obj(
        identifier, name_node, top_name_obj, prog_scope
    )


def find_declaration_by_scope_obj(
    identifier: ast.Identifier,
    name_node: ast.Name,
    top_name_obj: scope.ScopeObject,
    prog_scope: scope.ProgramScope,
) -> Optional[scope.ScopeObject]:
    if not identifier or not name_node or not top_name_obj or not prog_scope:
        return None
    index = identifier.name_nodes.index(name_node)
    i = 0
    obj = top_name_obj
    while i < index:
        i = i + 1
        obj = find_inner_name(obj, identifier.name_nodes[i], prog_scope)
        if not obj:
            return None
    return obj


def find_declaration_obj_by_pos_and_name(
    pos: ast.Position, name: str, prog_scope: scope.ProgramScope
) -> Optional[scope.ScopeObject]:
    if not pos or not pos.is_valid() or not name or not prog_scope:
        return None
    inner_most = prog_scope.main_scope.inner_most(pos)
    if not inner_most or not inner_most.elems:
        return None
    scope_obj = inner_most.elems.get(name)
    if scope_obj is not None:
        return scope_obj
    # 1. search through the parent schema scope tree
    parent_scope = inner_most.get_parent_schema_scope(prog_scope)
    while parent_scope is not None:
        scope_obj = parent_scope.elems.get(name)
        if scope_obj is not None:
            return scope_obj
        parent_scope = parent_scope.get_parent_schema_scope(prog_scope)
    # 2. search through the enclosing scope tree
    while inner_most is not None:
        scope_obj = inner_most.elems.get(name)
        if scope_obj is not None:
            return scope_obj
        inner_most = inner_most.get_enclosing_scope()
    return None


def find_inner_name(
    out_name_obj: scope.ScopeObject,
    inner_name: ast.Name,
    prog_scope: scope.ProgramScope,
) -> Optional[scope.ScopeObject]:
    if not out_name_obj or not inner_name or not prog_scope:
        return None
    if out_name_obj.type.type_kind() == KCLTypeKind.SchemaKind:
        return find_attr_by_name(inner_name.value, out_name_obj.type, prog_scope)
    if out_name_obj.type.type_kind() == KCLTypeKind.ModuleKind:
        out_type = cast(KCLModuleTypeObject, out_name_obj.type)
        if out_type.is_user_module:
            pkg_scope = prog_scope.scope_map.get(out_type.pkgpath)
            return pkg_scope.elems.get(inner_name.value) if pkg_scope else None
    if out_name_obj.type.type_kind() == KCLTypeKind.SchemaDefKind:
        out_type = cast(KCLSchemaDefTypeObject, out_name_obj.type)
        return find_attr_by_name(inner_name.value, out_type.schema_type, prog_scope)


def find_attr_by_name(
    attr_name: str, schema_type: KCLSchemaTypeObject, prog_scope: scope.ProgramScope
) -> Optional[scope.ScopeObject]:
    while schema_type:
        if attr_name in schema_type.attr_list:
            # todo: support jump to schema index signature
            pkg_scope = prog_scope.scope_map.get(schema_type.pkgpath)
            schema_scope = pkg_scope.search_child_scope_by_name(schema_type.name)
            return schema_scope.elems.get(attr_name) if schema_scope else None
        schema_type = schema_type.base
    return None



# Code generated by protoc-gen-protorpc-py. DO NOT EDIT.
#
# plugin: https://github.com/chai2010/protorpc/protoc-gen-plugin
# plugin: https://github.com/chai2010/protorpc-py/protoc-gen-protorpc-py
#
# source: gpyrpc.proto

import abc
import sys
import typing

from google.protobuf import message as _message

from .protorpc import ServiceMeta as _ServiceMeta
from .protorpc import Server as _Server

from .gpyrpc_pb2 import (
    Complete_Args,
    Complete_Result,
    DocumentSymbol_Args,
    DocumentSymbol_Result,
    EvalCode_Args,
    EvalCode_Result,
    ExecProgram_Args,
    ExecProgram_Result,
    FormatCode_Args,
    FormatCode_Result,
    FormatPath_Args,
    FormatPath_Result,
    GetSchemaTypeMapping_Args,
    GetSchemaTypeMapping_Result,
    GetSchemaType_Args,
    GetSchemaType_Result,
    GoToDef_Args,
    GoToDef_Result,
    Hover_Args,
    Hover_Result,
    LintPath_Args,
    LintPath_Result,
    ListDepFiles_Args,
    ListDepFiles_Result,
    ListMethod_Args,
    ListMethod_Result,
    LoadSettingsFiles_Args,
    LoadSettingsFiles_Result,
    OverrideFile_Args,
    OverrideFile_Result,
    ParseFile_AST_Args,
    ParseFile_AST_Result,
    ParseFile_LarkTree_Args,
    ParseFile_LarkTree_Result,
    ParseProgram_AST_Args,
    ParseProgram_AST_Result,
    Ping_Args,
    Ping_Result,
    ResetPlugin_Args,
    ResetPlugin_Result,
    ResolveCode_Args,
    ResolveCode_Result,
    SpliceCode_Args,
    SpliceCode_Result,
    ValidateCode_Args,
    ValidateCode_Result,
)


class BuiltinService(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Ping(self, args: Ping_Args) -> Ping_Result:
        pass

    @abc.abstractmethod
    def ListMethod(self, args: ListMethod_Args) -> ListMethod_Result:
        pass


class KclvmService(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Ping(self, args: Ping_Args) -> Ping_Result:
        pass

    @abc.abstractmethod
    def ParseFile_LarkTree(
        self, args: ParseFile_LarkTree_Args
    ) -> ParseFile_LarkTree_Result:
        pass

    @abc.abstractmethod
    def ParseFile_AST(self, args: ParseFile_AST_Args) -> ParseFile_AST_Result:
        pass

    @abc.abstractmethod
    def ParseProgram_AST(self, args: ParseProgram_AST_Args) -> ParseProgram_AST_Result:
        pass

    @abc.abstractmethod
    def ExecProgram(self, args: ExecProgram_Args) -> ExecProgram_Result:
        pass

    @abc.abstractmethod
    def ResetPlugin(self, args: ResetPlugin_Args) -> ResetPlugin_Result:
        pass

    @abc.abstractmethod
    def FormatCode(self, args: FormatCode_Args) -> FormatCode_Result:
        pass

    @abc.abstractmethod
    def FormatPath(self, args: FormatPath_Args) -> FormatPath_Result:
        pass

    @abc.abstractmethod
    def LintPath(self, args: LintPath_Args) -> LintPath_Result:
        pass

    @abc.abstractmethod
    def OverrideFile(self, args: OverrideFile_Args) -> OverrideFile_Result:
        pass

    @abc.abstractmethod
    def EvalCode(self, args: EvalCode_Args) -> EvalCode_Result:
        pass

    @abc.abstractmethod
    def ResolveCode(self, args: ResolveCode_Args) -> ResolveCode_Result:
        pass

    @abc.abstractmethod
    def GetSchemaType(self, args: GetSchemaType_Args) -> GetSchemaType_Result:
        pass

    @abc.abstractmethod
    def GetSchemaTypeMapping(
        self, args: GetSchemaTypeMapping_Args
    ) -> GetSchemaTypeMapping_Result:
        pass

    @abc.abstractmethod
    def ValidateCode(self, args: ValidateCode_Args) -> ValidateCode_Result:
        pass

    @abc.abstractmethod
    def SpliceCode(self, args: SpliceCode_Args) -> SpliceCode_Result:
        pass

    @abc.abstractmethod
    def Complete(self, args: Complete_Args) -> Complete_Result:
        pass

    @abc.abstractmethod
    def GoToDef(self, args: GoToDef_Args) -> GoToDef_Result:
        pass

    @abc.abstractmethod
    def DocumentSymbol(self, args: DocumentSymbol_Args) -> DocumentSymbol_Result:
        pass

    @abc.abstractmethod
    def Hover(self, args: Hover_Args) -> Hover_Result:
        pass

    @abc.abstractmethod
    def ListDepFiles(self, args: ListDepFiles_Args) -> ListDepFiles_Result:
        pass

    @abc.abstractmethod
    def LoadSettingsFiles(
        self, args: LoadSettingsFiles_Args
    ) -> LoadSettingsFiles_Result:
        pass


class BuiltinService_Meta(_ServiceMeta):
    def __init__(self, instance: BuiltinService):
        super().__init__()
        self._instance = instance

    def get_service_name(self) -> str:
        return "BuiltinService"

    def get_method_list(self) -> typing.List[str]:
        return [
            "Ping",
            "ListMethod",
        ]

    def create_method_req_message(self, method: str) -> _message.Message:
        if method in ["Ping", "BuiltinService.Ping"]:
            return Ping_Args()
        if method in ["ListMethod", "BuiltinService.ListMethod"]:
            return ListMethod_Args()
        raise Exception(f"unknown method: {method}")

    def create_method_resp_message(self, method: str) -> _message.Message:
        if method in ["Ping", "BuiltinService.Ping"]:
            return Ping_Result()
        if method in ["ListMethod", "BuiltinService.ListMethod"]:
            return ListMethod_Result()
        raise Exception(f"unknown method: {method}")

    def get_service_instance(self) -> _message.Message:
        return typing.cast(_message.Message, self._instance)

    def call_method(self, method: str, req: _message.Message) -> _message.Message:
        if method in ["Ping", "BuiltinService.Ping"]:
            return self._instance.Ping(req)
        if method in ["ListMethod", "BuiltinService.ListMethod"]:
            return self._instance.ListMethod(req)
        raise Exception(f"unknown method: {method}")


class KclvmService_Meta(_ServiceMeta):
    def __init__(self, instance: KclvmService):
        super().__init__()
        self._instance = instance

    def get_service_name(self) -> str:
        return "KclvmService"

    def get_method_list(self) -> typing.List[str]:
        return [
            "Ping",
            "ParseFile_LarkTree",
            "ParseFile_AST",
            "ParseProgram_AST",
            "ExecProgram",
            "ResetPlugin",
            "FormatCode",
            "FormatPath",
            "LintPath",
            "OverrideFile",
            "EvalCode",
            "ResolveCode",
            "GetSchemaType",
            "GetSchemaTypeMapping",
            "ValidateCode",
            "SpliceCode",
            "Complete",
            "GoToDef",
            "DocumentSymbol",
            "Hover",
            "ListDepFiles",
            "LoadSettingsFiles",
        ]

    def create_method_req_message(self, method: str) -> _message.Message:
        if method in ["Ping", "KclvmService.Ping"]:
            return Ping_Args()
        if method in ["ParseFile_LarkTree", "KclvmService.ParseFile_LarkTree"]:
            return ParseFile_LarkTree_Args()
        if method in ["ParseFile_AST", "KclvmService.ParseFile_AST"]:
            return ParseFile_AST_Args()
        if method in ["ParseProgram_AST", "KclvmService.ParseProgram_AST"]:
            return ParseProgram_AST_Args()
        if method in ["ExecProgram", "KclvmService.ExecProgram"]:
            return ExecProgram_Args()
        if method in ["ResetPlugin", "KclvmService.ResetPlugin"]:
            return ResetPlugin_Args()
        if method in ["FormatCode", "KclvmService.FormatCode"]:
            return FormatCode_Args()
        if method in ["FormatPath", "KclvmService.FormatPath"]:
            return FormatPath_Args()
        if method in ["LintPath", "KclvmService.LintPath"]:
            return LintPath_Args()
        if method in ["OverrideFile", "KclvmService.OverrideFile"]:
            return OverrideFile_Args()
        if method in ["EvalCode", "KclvmService.EvalCode"]:
            return EvalCode_Args()
        if method in ["ResolveCode", "KclvmService.ResolveCode"]:
            return ResolveCode_Args()
        if method in ["GetSchemaType", "KclvmService.GetSchemaType"]:
            return GetSchemaType_Args()
        if method in ["GetSchemaTypeMapping", "KclvmService.GetSchemaTypeMapping"]:
            return GetSchemaTypeMapping_Args()
        if method in ["ValidateCode", "KclvmService.ValidateCode"]:
            return ValidateCode_Args()
        if method in ["SpliceCode", "KclvmService.SpliceCode"]:
            return SpliceCode_Args()
        if method in ["Complete", "KclvmService.Complete"]:
            return Complete_Args()
        if method in ["GoToDef", "KclvmService.GoToDef"]:
            return GoToDef_Args()
        if method in ["DocumentSymbol", "KclvmService.DocumentSymbol"]:
            return DocumentSymbol_Args()
        if method in ["Hover", "KclvmService.Hover"]:
            return Hover_Args()
        if method in ["ListDepFiles", "KclvmService.ListDepFiles"]:
            return ListDepFiles_Args()
        if method in ["LoadSettingsFiles", "KclvmService.LoadSettingsFiles"]:
            return LoadSettingsFiles_Args()
        raise Exception(f"unknown method: {method}")

    def create_method_resp_message(self, method: str) -> _message.Message:
        if method in ["Ping", "KclvmService.Ping"]:
            return Ping_Result()
        if method in ["ParseFile_LarkTree", "KclvmService.ParseFile_LarkTree"]:
            return ParseFile_LarkTree_Result()
        if method in ["ParseFile_AST", "KclvmService.ParseFile_AST"]:
            return ParseFile_AST_Result()
        if method in ["ParseProgram_AST", "KclvmService.ParseProgram_AST"]:
            return ParseProgram_AST_Result()
        if method in ["ExecProgram", "KclvmService.ExecProgram"]:
            return ExecProgram_Result()
        if method in ["ResetPlugin", "KclvmService.ResetPlugin"]:
            return ResetPlugin_Result()
        if method in ["FormatCode", "KclvmService.FormatCode"]:
            return FormatCode_Result()
        if method in ["FormatPath", "KclvmService.FormatPath"]:
            return FormatPath_Result()
        if method in ["LintPath", "KclvmService.LintPath"]:
            return LintPath_Result()
        if method in ["OverrideFile", "KclvmService.OverrideFile"]:
            return OverrideFile_Result()
        if method in ["EvalCode", "KclvmService.EvalCode"]:
            return EvalCode_Result()
        if method in ["ResolveCode", "KclvmService.ResolveCode"]:
            return ResolveCode_Result()
        if method in ["GetSchemaType", "KclvmService.GetSchemaType"]:
            return GetSchemaType_Result()
        if method in ["GetSchemaTypeMapping", "KclvmService.GetSchemaTypeMapping"]:
            return GetSchemaTypeMapping_Result()
        if method in ["ValidateCode", "KclvmService.ValidateCode"]:
            return ValidateCode_Result()
        if method in ["SpliceCode", "KclvmService.SpliceCode"]:
            return SpliceCode_Result()
        if method in ["Complete", "KclvmService.Complete"]:
            return Complete_Result()
        if method in ["GoToDef", "KclvmService.GoToDef"]:
            return GoToDef_Result()
        if method in ["DocumentSymbol", "KclvmService.DocumentSymbol"]:
            return DocumentSymbol_Result()
        if method in ["Hover", "KclvmService.Hover"]:
            return Hover_Result()
        if method in ["ListDepFiles", "KclvmService.ListDepFiles"]:
            return ListDepFiles_Result()
        if method in ["LoadSettingsFiles", "KclvmService.LoadSettingsFiles"]:
            return LoadSettingsFiles_Result()
        raise Exception(f"unknown method: {method}")

    def get_service_instance(self) -> _message.Message:
        return typing.cast(_message.Message, self._instance)

    def call_method(self, method: str, req: _message.Message) -> _message.Message:
        if method in ["Ping", "KclvmService.Ping"]:
            return self._instance.Ping(req)
        if method in ["ParseFile_LarkTree", "KclvmService.ParseFile_LarkTree"]:
            return self._instance.ParseFile_LarkTree(req)
        if method in ["ParseFile_AST", "KclvmService.ParseFile_AST"]:
            return self._instance.ParseFile_AST(req)
        if method in ["ParseProgram_AST", "KclvmService.ParseProgram_AST"]:
            return self._instance.ParseProgram_AST(req)
        if method in ["ExecProgram", "KclvmService.ExecProgram"]:
            return self._instance.ExecProgram(req)
        if method in ["ResetPlugin", "KclvmService.ResetPlugin"]:
            return self._instance.ResetPlugin(req)
        if method in ["FormatCode", "KclvmService.FormatCode"]:
            return self._instance.FormatCode(req)
        if method in ["FormatPath", "KclvmService.FormatPath"]:
            return self._instance.FormatPath(req)
        if method in ["LintPath", "KclvmService.LintPath"]:
            return self._instance.LintPath(req)
        if method in ["OverrideFile", "KclvmService.OverrideFile"]:
            return self._instance.OverrideFile(req)
        if method in ["EvalCode", "KclvmService.EvalCode"]:
            return self._instance.EvalCode(req)
        if method in ["ResolveCode", "KclvmService.ResolveCode"]:
            return self._instance.ResolveCode(req)
        if method in ["GetSchemaType", "KclvmService.GetSchemaType"]:
            return self._instance.GetSchemaType(req)
        if method in ["GetSchemaTypeMapping", "KclvmService.GetSchemaTypeMapping"]:
            return self._instance.GetSchemaTypeMapping(req)
        if method in ["ValidateCode", "KclvmService.ValidateCode"]:
            return self._instance.ValidateCode(req)
        if method in ["SpliceCode", "KclvmService.SpliceCode"]:
            return self._instance.SpliceCode(req)
        if method in ["Complete", "KclvmService.Complete"]:
            return self._instance.Complete(req)
        if method in ["GoToDef", "KclvmService.GoToDef"]:
            return self._instance.GoToDef(req)
        if method in ["DocumentSymbol", "KclvmService.DocumentSymbol"]:
            return self._instance.DocumentSymbol(req)
        if method in ["Hover", "KclvmService.Hover"]:
            return self._instance.Hover(req)
        if method in ["ListDepFiles", "KclvmService.ListDepFiles"]:
            return self._instance.ListDepFiles(req)
        if method in ["LoadSettingsFiles", "KclvmService.LoadSettingsFiles"]:
            return self._instance.LoadSettingsFiles(req)
        raise Exception(f"unknown method: {method}")


class BuiltinService_Server:
    def __init__(self, instance: BuiltinService):
        self.instance = instance

    def run(self, *, stdin=sys.stdin, stdout=sys.stdout):
        rpc_server = _Server()
        rpc_server.register_service(BuiltinService_Meta(self.instance))
        rpc_server.run(stdin=stdin, stdout=stdout)


class KclvmService_Server:
    def __init__(self, instance: KclvmService):
        self.instance = instance

    def run(self, *, stdin=sys.stdin, stdout=sys.stdout):
        rpc_server = _Server()
        rpc_server.register_service(KclvmService_Meta(self.instance))
        rpc_server.run(stdin=stdin, stdout=stdout)



# Copyright 2021 The KCL Authors. All rights reserved.

from __future__ import annotations  # for python 3.7

from typing import List
from dataclasses import dataclass, field

# protobuf:
#
# message Tree {
#     string type = 1;
#     string token_value = 2;
#     repeated Tree children = 3;
#
#     int32 line = 101;
#     int32 column = 102;
#     int32 end_line = 103;
#     int32 end_column = 104;
# }


@dataclass
class Tree:
    type: str = ""
    token_value: str = ""
    children: List[Tree] = field(default_factory=lambda: [])

    line: int = 0
    column: int = 0
    end_line: int = 0
    end_column: int = 0



# Copyright 2020 The KCL Authors. All rights reserved.

import typing
from abc import abstractmethod

import kclvm.kcl.ast as ast


class TreeWalker:
    """The TreeWalk class can be used as a superclass in order
    to walk an AST or similar tree.

    This class is meant to be subclassed, with the subclass adding walker
    methods.

    Per default the walker functions for the nodes are ``'walk_'`` +
    class name of the node.  So a `expr_stmt` node visit function would
    be `walk_expr_stmt`.  This behavior can be changed by overriding
    the `walk` method.  If no walker function exists for a node
    (return value `None`) the `generic_walker` walker is used instead.
    """

    _WALK_FUNCTION_PREFIX = "walk_"

    def walk(self, node) -> None:
        """Visit a node."""
        method = self._WALK_FUNCTION_PREFIX + self.get_node_name(node)
        walker = getattr(self, method, self.generic_walk)
        return walker(node)

    @abstractmethod
    def generic_walk(self, node):
        """Called if no explicit walker function exists for a node."""
        pass

    @abstractmethod
    def get_node_name(self, node):
        """Called if no explicit walker function exists for a node."""
        pass


# -----------------------------------------------------------------------------
# WalkTree
# -----------------------------------------------------------------------------


def _WalkTree(
    t: ast.AST, walk_fn: typing.Callable[[ast.AST], typing.Optional[typing.Callable]]
):
    if t and callable(walk_fn):
        walk_fn = walk_fn(t)
    if (not t) or (not callable(walk_fn)):
        return

    if isinstance(t, ast.TypeAliasStmt):
        return

    if isinstance(t, ast.ExprStmt):
        node = typing.cast(ast.ExprStmt, t)
        for x in node.exprs or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.UnificationStmt):
        node = typing.cast(ast.UnificationStmt, t)
        if node.target:
            _WalkTree(node.target, walk_fn)
        if node.value:
            _WalkTree(node.value, walk_fn)
        return

    if isinstance(t, ast.AssignStmt):
        node = typing.cast(ast.AssignStmt, t)
        for x in node.targets or []:
            _WalkTree(x, walk_fn)
        if node.value:
            _WalkTree(node.value, walk_fn)
        return

    if isinstance(t, ast.AugAssignStmt):
        node = typing.cast(ast.AugAssignStmt, t)
        if node.target:
            _WalkTree(node.target, walk_fn)
        if node.value:
            _WalkTree(node.value, walk_fn)
        return

    if isinstance(t, ast.AssertStmt):
        node = typing.cast(ast.AssertStmt, t)
        if node.test:
            _WalkTree(node.test, walk_fn)
        if node.if_cond:
            _WalkTree(node.if_cond, walk_fn)
        if node.msg:
            _WalkTree(node.msg, walk_fn)
        return

    if isinstance(t, ast.IfStmt):
        node = typing.cast(ast.IfStmt, t)
        if node.cond:
            _WalkTree(node.cond, walk_fn)
        for x in node.body or []:
            _WalkTree(x, walk_fn)
        for x in node.elif_cond or []:
            _WalkTree(x, walk_fn)
        for if_body in node.elif_body or []:
            for x in if_body:
                _WalkTree(x, walk_fn)
        for x in node.else_body or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.ImportStmt):
        return

    if isinstance(t, ast.SchemaIndexSignature):
        node = typing.cast(ast.SchemaIndexSignature, t)
        if node.value:
            _WalkTree(node.value, walk_fn)
        return

    if isinstance(t, ast.SchemaAttr):
        node = typing.cast(ast.SchemaAttr, t)
        if node.value:
            _WalkTree(node.value, walk_fn)
        for x in node.decorators or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.SchemaStmt):
        node = typing.cast(ast.SchemaStmt, t)
        if node.parent_name:
            _WalkTree(node.parent_name, walk_fn)
        if node.args:
            _WalkTree(node.args, walk_fn)
        if node.for_host_name:
            _WalkTree(node.for_host_name, walk_fn)
        if node.index_signature:
            _WalkTree(node.index_signature, walk_fn)
        for x in node.mixins or []:
            _WalkTree(x, walk_fn)
        for x in node.body or []:
            _WalkTree(x, walk_fn)
        for x in node.decorators or []:
            _WalkTree(x, walk_fn)
        for x in node.checks or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.RuleStmt):
        node = typing.cast(ast.RuleStmt, t)
        for x in node.parent_rules or []:
            _WalkTree(x, walk_fn)
        for x in node.decorators or []:
            _WalkTree(x, walk_fn)
        if node.args:
            _WalkTree(node.args, walk_fn)
        if node.for_host_name:
            _WalkTree(node.for_host_name, walk_fn)
        for x in node.checks or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.IfExpr):
        node = typing.cast(ast.IfExpr, t)
        if node.body:
            _WalkTree(node.body, walk_fn)
        if node.cond:
            _WalkTree(node.cond, walk_fn)
        if node.orelse:
            _WalkTree(node.orelse, walk_fn)
        return

    if isinstance(t, ast.UnaryExpr):
        node = typing.cast(ast.UnaryExpr, t)
        if node.operand:
            _WalkTree(node.operand, walk_fn)
        return

    if isinstance(t, ast.BinaryExpr):
        node = typing.cast(ast.BinaryExpr, t)
        if node.left:
            _WalkTree(node.left, walk_fn)
        if node.right:
            _WalkTree(node.right, walk_fn)
        return

    if isinstance(t, ast.SelectorExpr):
        node = typing.cast(ast.SelectorExpr, t)
        if node.value:
            _WalkTree(node.value, walk_fn)
        if node.attr:
            _WalkTree(node.attr, walk_fn)
        return

    if isinstance(t, ast.CallExpr):
        node = typing.cast(ast.CallExpr, t)
        if node.func:
            _WalkTree(node.func, walk_fn)
        for x in node.args or []:
            _WalkTree(x, walk_fn)
        for x in node.keywords or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.ParenExpr):
        node = typing.cast(ast.ParenExpr, t)
        _WalkTree(node.expr, walk_fn)
        return

    if isinstance(t, ast.QuantExpr):
        node = typing.cast(ast.QuantExpr, t)
        for var in node.variables:
            _WalkTree(var, walk_fn)
        if node.target:
            _WalkTree(node.target, walk_fn)
        if node.test:
            _WalkTree(node.test, walk_fn)
        if node.if_cond:
            _WalkTree(node.if_cond, walk_fn)
        return

    if isinstance(t, ast.ListExpr):
        node = typing.cast(ast.ListExpr, t)
        for x in node.elts or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.ListComp):
        node = typing.cast(ast.ListComp, t)
        if node.elt:
            _WalkTree(node.elt, walk_fn)
        for x in node.generators or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.ListIfItemExpr):
        node = typing.cast(ast.ListIfItemExpr, t)
        if node.if_cond:
            _WalkTree(node.if_cond, walk_fn)
        for expr in node.exprs or []:
            _WalkTree(expr, walk_fn)
        if node.orelse:
            _WalkTree(node.orelse, walk_fn)
        return

    if isinstance(t, ast.StarredExpr):
        node = typing.cast(ast.StarredExpr, t)
        if node.value:
            _WalkTree(node.value, walk_fn)
        return

    if isinstance(t, ast.ConfigIfEntryExpr):
        node = typing.cast(ast.ConfigIfEntryExpr, t)
        if node.if_cond:
            _WalkTree(node.if_cond, walk_fn)
        for key in node.keys or []:
            _WalkTree(key, walk_fn)
        for value in node.values or []:
            _WalkTree(value, walk_fn)
        if node.orelse:
            _WalkTree(node.orelse, walk_fn)
        return

    if isinstance(t, ast.ConfigExpr):
        node = typing.cast(ast.ConfigExpr, t)
        for x in node.keys or []:
            _WalkTree(x, walk_fn)
        for x in node.values or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.DictComp):
        node = typing.cast(ast.DictComp, t)
        if node.key:
            _WalkTree(node.key, walk_fn)
        if node.value:
            _WalkTree(node.value, walk_fn)
        for x in node.generators or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.CompClause):
        node = typing.cast(ast.CompClause, t)
        for x in node.targets or []:
            _WalkTree(x, walk_fn)
        if node.iter:
            _WalkTree(node.iter, walk_fn)
        for x in node.ifs or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.SchemaExpr):
        node = typing.cast(ast.SchemaExpr, t)
        if node.name:
            _WalkTree(node.name, walk_fn)
        for x in node.args or []:
            _WalkTree(x, walk_fn)
        for x in node.kwargs or []:
            _WalkTree(x, walk_fn)
        if node.config:
            _WalkTree(node.config, walk_fn)
        return

    if isinstance(t, ast.CheckExpr):
        node = typing.cast(ast.CheckExpr, t)
        if node.test:
            _WalkTree(node.test, walk_fn)
        if node.if_cond:
            _WalkTree(node.if_cond, walk_fn)
        if node.msg:
            _WalkTree(node.msg, walk_fn)
        return

    if isinstance(t, ast.LambdaExpr):
        node = typing.cast(ast.LambdaExpr, t)
        if node.args:
            _WalkTree(node.args, walk_fn)
        for x in node.body or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.Decorator):
        node = typing.cast(ast.Decorator, t)
        if node.name:
            _WalkTree(node.name, walk_fn)
        if node.args:
            _WalkTree(node.args, walk_fn)
        return

    if isinstance(t, ast.Subscript):
        node = typing.cast(ast.Subscript, t)
        if node.value:
            _WalkTree(node.value, walk_fn)
        if node.index:
            _WalkTree(node.index, walk_fn)
        if node.lower:
            _WalkTree(node.lower, walk_fn)
        if node.upper:
            _WalkTree(node.upper, walk_fn)
        if node.step:
            _WalkTree(node.step, walk_fn)
        return

    if isinstance(t, ast.Keyword):
        node = typing.cast(ast.Keyword, t)
        if node.arg:
            _WalkTree(node.arg, walk_fn)
        if node.value:
            _WalkTree(node.value, walk_fn)
        return

    if isinstance(t, ast.Arguments):
        node = typing.cast(ast.Arguments, t)
        for x in node.args or []:
            _WalkTree(x, walk_fn)
        for x in node.defaults or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.Compare):
        node = typing.cast(ast.Compare, t)
        if node.left:
            _WalkTree(node.left, walk_fn)
        for x in node.comparators or []:
            _WalkTree(x, walk_fn)
        return

    if isinstance(t, ast.Identifier):
        _ = typing.cast(ast.Identifier, t)
        return

    if isinstance(t, ast.Literal):
        _ = typing.cast(ast.Literal, t)
        return
    if isinstance(t, ast.JoinedString):
        node = typing.cast(ast.JoinedString, t)
        for x in node.values or []:
            _WalkTree(x, walk_fn)
        return
    if isinstance(t, ast.FormattedValue):
        node = typing.cast(ast.FormattedValue, t)
        if node.value:
            _WalkTree(node.value, walk_fn)
        return

    if isinstance(t, ast.Module):
        node = typing.cast(ast.Module, t)
        for x in node.body or []:
            _WalkTree(x, walk_fn)
        return

    assert False, f"_WalkTree: t = {t}"


def WalkTree(m: ast.AST, walk_fn: typing.Callable[[ast.AST], None]):
    _WalkTree(m, walk_fn)



import sys
import argparse
import traceback

from kclvm.tools.format.format import kcl_fmt


def kcl_format_main():
    """KCL format tool CLI entry point"""
    parser = argparse.ArgumentParser(prog="kcl-fmt")
    parser.add_argument(
        dest="file",
        type=str,
        help="Input file or path name for formatting",
    )
    parser.add_argument(
        "-R",
        "--recursive",
        dest="recursive",
        help="Iterate through subdirectories recursively",
        action="store_true",
        required=False,
    )
    parser.add_argument(
        "-w",
        "--std-output",
        dest="std_output",
        help="Whether to output format to stdout",
        action="store_true",
        required=False,
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug_mode",
        help="Run in debug mode (for developers only)",
        action="store_true",
        required=False,
    )
    args = parser.parse_args()
    if len(sys.argv) == 1:
        parser.print_help(sys.stdout)
        sys.exit(0)

    try:
        kcl_fmt(args.file, is_stdout=args.std_output, recursively=args.recursive)
    except Exception as err:
        if args.debug_mode:
            print(traceback.format_exc())
        else:
            print(f"{err}")


if __name__ == "__main__":
    kcl_format_main()



# Copyright 2021 The KCL Authors. All rights reserved.

from .vfs import (
    LoadPkgCache,
    SavePkgCache,
    LoadMainPkgCache,
    SaveMainPkgCache,
    LoadBytecodeCache,
    SaveBytecodeCache,
    IsAbsPkgPath,
    IsRelPkgPath,
    FixImportPath,
    CacheOption,
    DEFAULT_CACHE_DIR,
    FST_CACHE_DIR,
)
from .kcl_mod import GetPkgRoot, MustGetPkgRoot, LoadModFile

__all__ = [
    "LoadPkgCache",
    "SavePkgCache",
    "LoadMainPkgCache",
    "SaveMainPkgCache",
    "LoadBytecodeCache",
    "SaveBytecodeCache",
    "IsAbsPkgPath",
    "IsRelPkgPath",
    "FixImportPath",
    "GetPkgRoot",
    "MustGetPkgRoot",
    "LoadModFile",
    "CacheOption",
    "DEFAULT_CACHE_DIR",
    "FST_CACHE_DIR",
]



import hashlib as _hashlib


def KMANGLED_md5(value: str, encoding: str = "utf-8") -> str:
    return _hashlib.md5(value.encode(encoding)).hexdigest()


def KMANGLED_sha1(value: str, encoding: str = "utf-8") -> str:
    return _hashlib.sha1(value.encode(encoding)).hexdigest()


def KMANGLED_sha224(value: str, encoding: str = "utf-8") -> str:
    return _hashlib.sha224(value.encode(encoding)).hexdigest()


def KMANGLED_sha256(value: str, encoding: str = "utf-8") -> str:
    return _hashlib.sha256(value.encode(encoding)).hexdigest()


def KMANGLED_sha384(value: str, encoding: str = "utf-8") -> str:
    return _hashlib.sha384(value.encode(encoding)).hexdigest()


def KMANGLED_sha512(value: str, encoding: str = "utf-8") -> str:
    return _hashlib.sha512(value.encode(encoding)).hexdigest()



from dataclasses import dataclass
from enum import Enum
from typing import Optional, List, Dict, Tuple

import kclvm.kcl.info as kcl_info

from kclvm.compiler.extension.builtin import BUILTIN_FUNCTIONS


class SymbolScope(Enum):
    """Symbol Scope

    Parameters
    ----------
    GLOBAL: str
        Top level variables including variables in IfStmt

    LOCAL: str
        Variables in schema context or in for comprehension

    BUILTIN: str
        Built-in variables e.g., `print`, `sum`,

    FREE: str
        Variables that need to be captured by closures

    INTERNAL: str
        Internal variables used only for record names e.g., `b` and `c` in `a.b.c`
    """

    GLOBAL = "GLOBAL"
    LOCAL = "LOCAL"
    BUILT_IN = "BUILT_IN"
    FREE = "FREE"
    INTERNAL = "INTERNAL"


@dataclass
class Symbol:
    """Symbol

    Parameters
    ----------
    name: str
        The symbol name

    index: int
        The symbol index in the symbol table

    scope:
        The symbol scope

    define_count:
        The number of times the symbol is defined or declared
    """

    name: str
    index: int
    scope: SymbolScope = None
    define_count: int = 0


@dataclass
class SymbolTable:
    """Symbol table

    Parameters
    ----------
    outer: SymbolTable
        The symbol table in the outer symbol scope
        e.g., The schema context symbol table outer is the top level scope symbol table.

    store: Dict[str, Symbol]
        The current scope symbol map to store all symbols

    free_symbols: List[Symbol]
        The free symbol list in the current scope

    num_definitions: int
        The total number of symbols
    """

    outer: "SymbolTable" = None
    store: Dict[str, Symbol] = None
    free_symbols: List[Symbol] = None
    num_definitions: int = 0

    def define(self, name: str, scope: SymbolScope = None) -> Tuple[Symbol, bool]:
        """
        Define a symbol named 'name' with 'scope' and put it into symbol table
        """
        assert isinstance(name, str)

        def default_scope():
            """The inner default symbol scope

            If outer is exist, return LOCAL scope, else return GLOBAL scope
            """
            return SymbolScope.GLOBAL if self.outer is None else SymbolScope.LOCAL

        symbol = Symbol(
            name=name, index=self.num_definitions, scope=scope or default_scope()
        )
        is_exist = False
        if (
            name in self.store
            and not kcl_info.isprivate_field(name)
            and self.store[name].scope == SymbolScope.GLOBAL
        ):
            is_exist = True
            self.num_definitions += 1
            return self.store[name], is_exist
        # Internal scope variable skip exist symbol
        if symbol.scope == SymbolScope.INTERNAL and name in self.store:
            pass
        else:
            self.store[name] = symbol
        self.num_definitions += 1
        return symbol, is_exist

    def define_builtin(self, name: str, index: int) -> Symbol:
        """Define a builtin function object"""
        if self.store is None or not isinstance(self.store, dict):
            raise Exception("Invalid symbol table store")
        symbol = Symbol(name=name, index=index)
        symbol.scope = SymbolScope.BUILT_IN
        self.store[name] = symbol
        return symbol

    def define_free(self, original: Symbol) -> Symbol:
        """
        Define a symbol named 'name' with free scope and put it into symbol table
        """
        self.free_symbols.append(original)
        self.store[original.name] = Symbol(
            name=original.name, index=original.index, scope=SymbolScope.FREE
        )
        return self.store[original.name]

    def register_builtins(self):
        """Register builtin functions into symbol table"""
        builtins = BUILTIN_FUNCTIONS
        for i, builtin in enumerate(builtins):
            self.define_builtin(builtin, i)
        return self

    def resolve(self, name: str) -> Optional[Symbol]:
        """Resolve a symbol named 'name'

        Search from the current scope, if not found, search from the symbol table of its outer
        """
        obj = self.store.get(name)
        if not obj and self.outer:
            obj = self.outer.resolve(name)
            if not obj:
                return obj
            if obj.scope == SymbolScope.GLOBAL or obj.scope == SymbolScope.BUILT_IN:
                return obj
            elif obj.scope == SymbolScope.INTERNAL:
                return None
            return self.define_free(obj)
        return obj

    def delete(self, name: str, scope: SymbolScope):
        """Delete name from the symbol table"""
        if (
            not name
            or not scope
            or name not in self.store
            or self.store[name].scope != scope
        ):
            return
        del self.store[name]

    # Static methods

    @staticmethod
    def new(outer=None, num=0):
        """New an empty symbol table"""
        return SymbolTable(
            outer=outer,
            store={},
            free_symbols=[],
            num_definitions=num,
        )

    @staticmethod
    def new_with_built_in(outer=None, num=0):
        """New a symbol table with all builtin functions"""
        return SymbolTable.new(outer, num).register_builtins()



import sys
from typing import List

from kclvm.tools.lint.reporters.base_reporter import BaseReporter
from kclvm.tools.lint.message.message import Message


class FileReporter(BaseReporter):
    def __init__(self, linter, output=None, encoding=None):
        self.name = "file_reporter"
        self.output_file = linter.config.output_path
        super().__init__(linter, output, encoding)

    def print_msg(self, msgs: List[Message] = None):
        assert self.output_file
        with open(self.output_file, "w") as f:
            current = sys.stdout
            sys.stdout = f
            for msg in msgs:
                print(msg)
                print()
            print("Check total {} files:".format(len(self.linter.file_list)))
            for k, v in self.linter.msgs_map.items():
                print("{:<8}{}: {}".format(v, k, self.linter.MSGS[k][1]))
            print(f"KCL Lint: {len(self.linter.msgs)} problems")
            sys.stdout = current



# Copyright 2021 The KCL Authors. All rights reserved.

from typing import Dict

from .parser import Type, KeyType

PROTOBUF_SCALAR_TYPE_TO_KCL_MAPPING: Dict[str, str] = {
    "int32": "int",
    "int64": "int",
    "uint32": "int",
    "uint64": "int",
    "double": "float",
    "float": "float",
    "fixed32": "int",
    "fixed64": "int",
    "sfixed32": "int",
    "sfixed64": "int",
    "bool": "bool",
    "string": "str",
    "bytes": "str",
}

KCL_SCALAR_TYPE_TO_PROTOBUF_TYPE_MAPPING: Dict[str, Type] = {
    "int": Type.INT64,
    "float": Type.DOUBLE,
    "bool": Type.BOOL,
    "str": Type.STRING,
}

KCL_SCALAR_TYPE_TO_PROTOBUF_KEY_TYPE_MAPPING: Dict[str, Type] = {
    "int": KeyType.INT64,
    "float": KeyType.INT64,
    "bool": KeyType.BOOL,
    "str": KeyType.STRING,
}



import typing
import pathlib

import kclvm.compiler.parser.parser as parser
import kclvm.kcl.ast as ast
import kclvm.kcl.types.scope as scope
import kclvm.kcl.error as kcl_error

from kclvm.kcl.types import ResolveProgram, ProgramScope, CheckConfig
from pygls.lsp.types.basic_structures import Location, Range, Position
from pygls.lsp.types.language_features.completion import (
    CompletionItemKind,
    CompletionItem,
)
from kclvm.api.object import KCLTypeKind
from kclvm.compiler.vfs.vfs import get_pkg_realpath_from_pkgpath as get_realpath

INNER_TYPE_2_COMPLETION_ITEM_KIND = {
    # any type
    KCLTypeKind.AnyKind: CompletionItemKind.Text,
    # base types
    KCLTypeKind.BoolKind: CompletionItemKind.Value,
    KCLTypeKind.IntKind: CompletionItemKind.Value,
    KCLTypeKind.FloatKind: CompletionItemKind.Value,
    KCLTypeKind.StrKind: CompletionItemKind.Value,
    # constants
    KCLTypeKind.NoneKind: CompletionItemKind.Constant,
    KCLTypeKind.BoolLitKind: CompletionItemKind.Constant,
    KCLTypeKind.FloatLitKind: CompletionItemKind.Constant,
    KCLTypeKind.IntLitKind: CompletionItemKind.Constant,
    KCLTypeKind.StrLitKind: CompletionItemKind.Constant,
    # schema
    KCLTypeKind.SchemaKind: CompletionItemKind.Struct,
    KCLTypeKind.SchemaDefKind: CompletionItemKind.Struct,
    # built in function
    KCLTypeKind.FuncKind: CompletionItemKind.Function,
}


def pos_to_node(
    pos: ast.Position, code: str = None
) -> (typing.Optional[ast.Program], typing.Optional[ast.AST]):
    if not pos or not pos.is_valid():
        return None, None
    prog = file_to_prog(pos.filename, code)
    if not prog:
        return None, None
    for module in prog.pkgs[ast.Program.MAIN_PKGPATH]:
        leaf_node = module.find_leaf_by_pos(pos)
        if leaf_node:
            return prog, leaf_node
    return prog, None


def pos_to_scope(
    pos: ast.Position, code: str = None
) -> (
    typing.Optional[ast.Program],
    typing.Optional[scope.Scope],
    typing.Optional[scope.Scope],
):
    if not pos or not pos.is_valid():
        return None, None, None
    prog, prog_scope = file_or_prog_to_scope(None, pos.filename, code)
    if not prog_scope or not prog_scope.main_scope:
        return None, None, None
    return prog, prog_scope.main_scope.inner_most(pos), prog_scope


def file_or_prog_to_scope(
    prog: typing.Optional[ast.Program] = None, file_path: str = None, code: str = None
) -> (typing.Optional[ast.Program], typing.Optional[ProgramScope]):
    prog = prog or file_to_prog(file_path, code)
    if not prog:
        return None, None
    no_raise = CheckConfig()
    no_raise.raise_err = False
    try:
        return prog, ResolveProgram(prog, config=no_raise)
    except Exception:
        return None, None


def file_to_prog(file_path: str, code: str = None) -> typing.Optional[ast.Program]:
    code_list = [code] if code is not None else []
    try:
        prog = parser.LoadProgram(
            file_path,
            mode=parser.ParseMode.Null,
            k_code_list=code_list,
            set_ast_parent=True,
        )
        return prog
    except kcl_error.KCLError:
        return None
    except Exception:
        return None


def file_to_ast(file_path: str, code: str = None) -> typing.Optional[ast.Module]:
    try:
        module = parser.ParseFile(file_path, code)
        return module
    except Exception:
        return None


def scope_obj_to_location(obj: scope.ScopeObject) -> typing.Optional[Location]:
    if obj and obj.check_pos_valid():
        if obj.node.type in [
            "SchemaStmt",
            "RuleStmt",
            "SchemaAttr",
            "SchemaIndexSignature",
        ]:
            return node_to_location(obj.node.name_node)
        if obj.node.type == "Identifier":
            return node_to_location(obj.node.name_nodes[0])
        if obj.node.type == "ImportStmt":
            return node_to_location(
                obj.node.as_name_node
                if obj.node.as_name_node
                else obj.node.path_nodes[-1]
            )
    return None


def node_to_location(node: ast.AST) -> typing.Optional[Location]:
    return (
        Location(
            uri=str(pathlib.Path(node.filename)),
            range=Range(
                start=kcl_pos_to_lsp_pos(node.pos),
                end=kcl_pos_to_lsp_pos(node.end_pos),
            ),
        )
        if node
        and node.pos
        and node.pos.is_valid()
        and node.end_pos
        and node.end_pos.is_valid()
        else None
    )


def kcl_pos_to_lsp_pos(pos: ast.Position) -> Position:
    return Position(line=pos.line - 1, character=pos.column - 1)


def scope_obj_to_completion_item(obj: scope.ScopeObject) -> CompletionItem:
    assert obj and obj.name and obj.type
    return CompletionItem(
        label=obj.name,
        kind=INNER_TYPE_2_COMPLETION_ITEM_KIND.get(obj.type.type_kind()),
    )


def pkgpath_to_location(root: str, pkgpath: str) -> typing.Optional[Location]:
    filepath = get_realpath_from_pkgpath(root, pkgpath)
    return file_to_location(filepath)


def file_to_location(filepath: str) -> typing.Optional[Location]:
    if not filepath or not is_kcl_file(filepath):
        return None
    return Location(uri=filepath, range=emptyRange())


def is_kcl_file(filepath: str) -> bool:
    return filepath.endswith(".k")


def emptyRange() -> Range:
    return Range(
        start=Position(line=0, character=0),
        end=Position(line=0, character=0),
    )


def get_realpath_from_pkgpath(root: str, pkgpath: str) -> str:
    return str(pathlib.Path(get_realpath(root, pkgpath)).absolute())
